# pylint: disable=try-except-raise

import io
import os
import sys
import queue
import signal
import multiprocessing as mp

from .misc    import ru_open
from .testing import sys_exit

# from
# http://stackoverflow.com/questions/1417631/python-code-to-daemonize-a-process


# ------------------------------------------------------------------------------
#
def daemonize(main=None, args=None, stdout=None, stderr=None, stdin=None,
              timeout=None):
    '''
    Create a daemon process and run the given method in it.   For that, do the
    UNIX double-fork magic, see Stevens' "Advanced Programming in the UNIX
    Environment" for details (ISBN 0201563177)

    The method will return the PID of the spawned damon process, or `None` on
    failure to create it.  stdout, stderr, stdin file names are interpreted by
    the daemon process, and are expected to be path names which can be opened
    and read / written in their respective capabilities.
    '''

    if main:
        assert callable(main)

    pid   = None
    pid_q = mp.Queue()

    # first fork
    try:
        f1_pid = os.fork()
        if  f1_pid > 0:

            # wait for daemon pid from second parent
            # TODO: timeout
            if timeout:
                try:
                    pid = pid_q.get(timeout=timeout)
                except queue.Empty as e:
                    raise RuntimeError('daemon startup timed out') from e

            else:
                pid = pid_q.get()

            if not pid:
                raise RuntimeError('daemon startup failed')

            # we are done...
            return pid

    except OSError as e:
        raise RuntimeError(
            'Fork failed: %d (%s)\n' % (e.errno, e.strerror)) from e

    except Exception as e:
        raise RuntimeError(
            'Failed to start daemon: %d (%s)\n' % (e.errno, e.strerror)) from e

    # decoupling from parent process group is disabled
    # (some launch methods, e.g., APRun, required to run within the same group)
    #    os.setsid()

    # second fork
    try:
        f2_pid = os.fork()

        if  f2_pid > 0:
            # communicate pid to first parent
            pid_q.put(f2_pid)

            # exit from second parent
            sys_exit(0)

    except OSError:
        pid_q.put(None)  # unblock parent
        sys_exit(1)

    # redirect standard file descriptors
    if stdin:
        try:
            si = ru_open(stdin, 'r')
            os.dup2(si.fileno(), sys.stdin.fileno())
        except io.UnsupportedOperation:
            sys.stdin = ru_open(stdin, 'r')

    if stdout:
        try:
            sys.stdout.flush()
            so = ru_open(stdout, 'a+')
            os.dup2(so.fileno(), sys.stdout.fileno())
        except io.UnsupportedOperation:
            sys.stdout = ru_open(stdout, 'a+')

    if stderr:
        try:
            sys.stderr.flush()
            se = ru_open(stderr, 'a+')
            os.dup2(se.fileno(), sys.stderr.fileno())
        except io.UnsupportedOperation:
            sys.stderr = ru_open(stderr, 'a+')

    if main:
        # we are successfully daemonized - run the workload and exit
        if args is None: main()
        else           : main(*args)
        sys_exit(0)

    else:
        # just return - the calling code will now continue daemonized
        return


# ------------------------------------------------------------------------------
#
class Daemon(object):
    '''
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, stdin='/dev/null',
                       stdout='/dev/null',
                       stderr='/dev/null',
                       target=None,
                       args=None):

        if target:
            assert callable(target)

        self.stdin   = stdin
        self.stdout  = stdout
        self.stderr  = stderr
        self.pid     = None
        self.target  = target
        self.args    = args


    # --------------------------------------------------------------------------
    #
    def start(self):

        # start the daemon, and in the demon process, run the workload
        self.pid = daemonize(main=self.run, args=self.args, stdin=self.stdin,
                                                            stdout=self.stdout,
                                                            stderr=self.stderr)
        return self.pid


    # --------------------------------------------------------------------------
    #
    def stop(self, pid=None):
        '''
        Stop the daemon.  If a pid is passed, then stop the daeon process with
        that pid.
        '''

        if not pid:
            pid = self.pid

        if not pid:
            raise RuntimeError('no pid - daemon not started, yet?')

        # Try killing the daemon process
        os.kill(pid, signal.SIGTERM)


    # --------------------------------------------------------------------------
    #
    def restart(self, pid=None):
        '''
        Stop the daemon and restart it.  If a pid is passed, then stop the daeon
        process with that pid and replace it with a daemon process represented
        by this class instance.  It is the caller's responsibility to ensure
        that this makes semantic sense.

        This method returns the pid of the new daemon process (see `start()`).
        '''

        self.stop(pid=pid)
        return self.start()


    # --------------------------------------------------------------------------
    #
    def run(self):
        '''
        You should override this method when you subclass Daemon and do not pass
        `target` to the class constructor.  This method will be called after the
        daemon process has been created by start() or restart().
        '''

        self.target(*self.args)


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import time
import threading
import contextlib
import weakref
from .singleton  import Singleton


# ------------------------------------------------------------------------------
#
TIMEOUT   = 0.1  # sleep between lock checks
READONLY  = 'ReadOnly'
READWRITE = 'ReadWrite'


# ------------------------------------------------------------------------------
#
class _Registry(object, metaclass=Singleton):
    '''
    This singleton registry class maintains a set of object instances, by some
    entity ID.  Consumers can acquire and release those instances, for
    `READONLY` or `READWRITE` activities.  An unlimited number of READONLY
    activities can be active at any time, on any given instance -- but if any
    `READWRITE` acquisition will block until no other instances are leased
    anymore, and any `READONLY` acquisition will block until an eventual release
    of a `READWRITE` lease::

        my_obj = MyClass(eid='my_id')
        print my_obj.id   # prints my_id
        ...

        import raducal.utils as  ru
        ro_instance_1 = ru.Registry.acquire('my_id', ru.READONLY)
        ro_instance_2 = ru.Registry.acquire('my_id', ru.READONLY)
        ...

        rw_instance_1 = ru.Registry.acquire('my_id', ru.READWRITE)

        # the last call would block until the following two calls have been
        # issued:
        ru.Registry.release('my_id')
        ru.Registry.release('my_id')

    The correct semantic functionality of this class depends on the careful
    release of unused instances.  Using a `with` statement is encouraged, and
    a `lease` context manager is provided for that purpose::

        with ru.Registry.lease(eid, ru.READONLY) as ro_instance:
            ro_instance.do_something()
            ro_instance.do_nothing  ()
            ro_instance.do_something()

        with ru.Registry.lease(eid, ru.READWRITE) as rw_instance:
            rw_instance.change_something

    The registry will thus ensure that a consumer will always see instances
    which are not changed by a third party over the scope of a lease.

    *Requirements*

    Registered object instances must fulfill two requirements:

    * they must be `lockable`, i.e. during `aquire` we can call `entity.lock()`
      on them to activate a `threading.RLock`, and during `release()` we can
      call `entity.unlock()` to deactivate it.  Instances will thus guaranteed
      to be locked during lease.

    * they must be `identifiable`, i.e. they must have an `id` property.
      Alternatively, an entity ID can be passed as optional parameter during
      registration -- all followup methods will require that eid.


    *Note* (for Troy as consumer of this registry, but applicable in general):

    It is in general not a favourable design to have large, all-visible
    registries in a multi-component software stack, as ownership of state
    transitions can easily become blurry, and as a registry can also become
    a performance bottleneck on frequent queries -- so why are we doing this?

    First of all, state transitions in Troy *are* blurry to some extent, as Troy
    aims to support a variety of state diagram transitions, and thus the order
    of transitions is not pre-defined (first derive overlay then schedule CUs,
    or vice versa?).  Also, the ownership of state transitions for a workload is
    not easily scoped (the :class:`Planner` will move a workload from
    `DESCRIBED` to `PLANNED`, the :class:`WorkloadManager` will move it from
    `PLANNED` to `TRANSLATED`, etc.  And, finally, we want to allow for
    re-scheduling, re-translation, re-planning etc, which would require us to
    pass control of a workload back and forth between different modules.
    Finally, this seems to be useful for inspection and introspection of Troy
    activities related to specific workload instances.

    In that context, a registry seems the (much) lesser of two devils: The
    registry class will allow to register workload and overlay instances, and to
    acquire/release control over them.  The module which acquires control needs
    to ascertain that the respective workload and overlay instances are in
    a usable state for that module -- the registry is neither interpreting nor
    enforcing any state model on the managed instances -- that is up to the
    leasing module.  Neither will the registry perform garbage collection --
    active unregistration will remove instances from the registry, but not
    enforce a deletion.

    This is a singleton class.  We assume that Workload and Overlay IDs are
    unique.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self):
        '''
        Create a new Registry instance.
        '''

        # make this instance lockable
        self.lock = threading.RLock()

        self._registry = dict()
        self._session  = None  # this will be set by Troy.submit_workload()


    # ------------------------------------------------------------------------------
    #
    @contextlib.contextmanager
    def lease(self, oid, mode=READWRITE):
        try:
            yield self.acquire(oid, mode)
        finally:
            self.release(oid)

    # --------------------------------------------------------------------------
    #
    def register(self, entity, eid=None):
        '''
        register a new object instance.
        '''

        # lock manager before checking/manipulating the registry
        with self.lock:

            # check if instance is lockable
            # FIXME: better inspection for ru.Lockables
            if  not hasattr(entity, '__enter__') or \
                not hasattr(entity, '__exit__' ) or \
                not hasattr(entity, 'lock'     ) or \
                not hasattr(entity, 'unlock'   ) or \
                not hasattr(entity, 'locked'   ) or \
                not hasattr(entity, '_locked'  ) or \
                not hasattr(entity, '_rlock'   ):
                raise TypeError('Registry only manages lockables')

            # check if instance is identifiable
            if  not eid:
                if  not hasattr(entity, 'id'):
                    raise TypeError('Registry only manages identifiables')
                eid = entity.id

            if  eid in self._registry:
                raise ValueError("'%s' is already registered" % eid)

            self._registry[eid] = {}
            self._registry[eid]['ro_leases'] = 0  # not leased
            self._registry[eid]['rw_leases'] = 0  # not leased
            self._registry[eid]['entity']    = weakref.ref(entity)


    # --------------------------------------------------------------------------
    #
    def acquire(self, eid, mode):
        '''
        temporarily relinquish control over the referenced identity to the
        caller.
        '''

        # sanity check
        if eid not in self._registry:
            raise KeyError("'%s' is not registered" % eid)

        # wait for the entity to be fee for the expected usage
        while True:

            # lock manager before checking/manipulating the registry
            with  self.lock:

                if  mode == READONLY:
                    # make sure we have no active write lease

                    if  self._registry[eid]['rw_leases']:
                        # need to wait
                        time.sleep(TIMEOUT)
                        continue

                    else:
                        # free for READONLY use
                        self._registry[eid]['ro_leases'] += 1
                        break

                if  mode == READWRITE:
                    # make sure we have no active read or write lease

                    if  self._registry[eid]['rw_leases'] or \
                        self._registry[eid]['ro_leases']:
                        # need to wait
                        time.sleep(TIMEOUT)
                        continue

                    else:
                        # free for READWRITE use
                        self._registry[eid]['rw_leases'] += 1
                        break

        # acquire entity lock
        entity = self._registry[eid]['entity']()

        if entity is None:
            raise KeyError("'%s' was deallocated" % eid)

        else:
            # all is well...
            entity.lock()
            return entity


    # --------------------------------------------------------------------------
    #
    def release(self, eid):
        '''
        relinquish the control over the referenced entity
        '''

        # sanity check
        if eid not in self._registry:
            raise KeyError("'%s' is not registered" % eid)

        # lock manager before checking/manipulating the registry
        with  self.lock:

            if  not self._registry[eid]['ro_leases'] and \
                not self._registry[eid]['rw_leases']:
                raise ValueError("'%s' was not acquired" % eid)

            # release entity lease
            if  self._registry[eid]['ro_leases']:
                self._registry[eid]['ro_leases'] -= 1

            elif self._registry[eid]['rw_leases']:
                self._registry[eid]['rw_leases'] -= 1

            # release entity lock
            entity = self._registry[eid]['entity']()

            if entity is None:
                raise KeyError("'%s' was deallocated" % eid)

            else:
                # all is well...
                entity.unlock()


    # --------------------------------------------------------------------------
    #
    def unregister(self, eid):
        '''
        remove the reference entity from the registry, but do not explicitly
        call the entity's destructor.  This will unlock the entity.
        '''

        # sanity check
        if eid not in self._registry:
            raise KeyError("'%s' is not registered" % eid)

        # lock manager before checking/manipulating the registry
        with self.lock:

            entity = self._registry[eid]['entity']()

            if entity is None:
                raise KeyError("'%s' was deallocated" % eid)

            # unlock entity
            while self._registry[eid]['ro_leases']:
                entity.unlock()
                self._registry[eid]['ro_leases'] -= 1

            while self._registry[eid]['rw_leases']:
                entity.unlock()
                self._registry[eid]['rw_leases'] -= 1

            # remove entity from registry, w/o a trace...
            del self._registry[eid]


# ------------------------------------------------------------------------------
#
# create a global registry instance
#
Registry = _Registry()


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import os
import sys
import glob
import pprint

from importlib  import util as imp

from .singleton import Singleton
from .logger    import Logger
from .misc      import as_list


# ------------------------------------------------------------------------------
#
class PluginBase(object):
    '''
    This class serves as base class for any plugin managed by the plugin handler
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, descr: dict) -> None:
        '''
        This constructor MUST be called by any inheriting implementation.
        '''

        self._plugin_descr = descr

    @property
    def plugin_type(self) -> str:
        return self._plugin_descr['type']

    @property
    def plugin_name(self) -> str:
        return self._plugin_descr['name']

    @property
    def plugin_class(self) -> str:
        return self._plugin_descr['class']

    @property
    def plugin_version(self) -> str:
        return self._plugin_descr['version']

    @property
    def plugin_description(self) -> str:
        return self._plugin_descr['description']


# ------------------------------------------------------------------------------
#
class PluginManager(object, metaclass=Singleton):
    '''
    The PluginManager manages plugins of specific types: the manager can search
    for installed plugins, list and describe plugins found, load plugins, and
    instantiate the plugin for further use.

    Example:

        # try to load the 'echo' plugin from the 'radical' namespace
        plugin_type = 'echo'

        pm = radical.utils.PluginManager('radical')

        for plugin_name in pm.list(plugin_type):
            print plugin_name
            print pm.describe(plugin_type, plugin_name)

        default_plugin = pm.load('echo', 'default')

        default_plugin.init_plugin('world')
        default_plugin.run()  # prints 'hello default world'


    The plugins are expected to follow a specific naming and coding schema to be
    recognized by the plugin manager.  The naming schema is:

        [namespace].plugins.[ptype].plugin_[ptype]_[pname].py

    i.e. for the example above: `radical.plugins.echo.plugin_echo_default.py`

    The plugin code consists of two parts:  a plugin description, and a plugin
    class.  The description is a module level dictionary named
    `PLUGIN_DESCRIPTION`, the plugin class must have a class constructor
    `__init__(*args, **kwargs)` to create plugin instances for further use.

    At this point, we leave the definition of the exact plugin signatures open,
    but expect that to be more strictly defined per plugin type in the future.

    Note that the PluginManager construction is, at this point, not considered
    thread safe.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, namespaces):
        '''
        namespace: name of module (plugins are expected in namespace/plugins/)
        '''

        # import here to avoid circular imports
        self._namespaces = as_list(namespaces)
        self._registry   = dict()
        self._seen       = list()
        self._log        = Logger('radical.utils')

        for namespace in self._namespaces:
            self.load_plugins(namespace, self._log)


    # --------------------------------------------------------------------------
    #
    def seen(self, pfile):

        if pfile in self._seen:
            return True
        else:
            self._seen.append(pfile)
            return False


    # --------------------------------------------------------------------------
    #
    def register(self, ptype, pname, pinfo):

        if ptype not in self._registry:
            self._registry[ptype] = dict()

        if pname not in self._registry[ptype]:
            self._registry[ptype][pname] = pinfo


    # --------------------------------------------------------------------------
    #
    def retrieve(self, ptype, pname):

        return self._registry.get(ptype, {}).get(pname)


    # --------------------------------------------------------------------------
    #
    def load_plugins(self, namespace, log):
        '''
        Load all plugins for the given namespace.  Previously loaded plugins
        are overloaded.
        '''

        self._log.info('loading plugins for namespace %s' % namespace)

        # search for plugins in all system module paths
        for spath in sys.path:

            # we only load plugins installed under the namespace hierarchy
            npath  = namespace.replace('.', '/')
            ppath  = '%s/%s/plugin*/' %  (spath, npath)
            pglob1 = '*/plugin_*.py'
            pglob2 = 'plugin_*.py'

            # we assume that all python sources in that location are
            # suitable plugins
            pfiles = glob.glob(ppath + pglob1) + glob.glob(ppath + pglob2)

            if not pfiles:
                continue

            for pfile in pfiles:

                # from the full plugin file name, derive a short name for more
                # useful logging, duplication checks etc. -- simply remove the
                # namespace path portion...
                if pfile.startswith(spath):
                    pshort = pfile[len(spath):]
                else:
                    pshort = pfile

                if self.seen(pshort):
                    continue

                try:
                    modname = '%s.plugins.%s.%s' % (namespace,
                                os.path.basename(os.path.dirname(pfile)),
                                os.path.splitext(os.path.basename(pfile))[0])
                    # now load the plugin proper
                    spec   = imp.spec_from_file_location(modname, pfile)
                    plugin = imp.module_from_spec(spec)
                    spec.loader.exec_module(plugin)

                    # get plugin details from description
                    ptype  = plugin.PLUGIN_DESCRIPTION.get('type',        None)
                    pname  = plugin.PLUGIN_DESCRIPTION.get('name',        None)
                    pclass = plugin.PLUGIN_DESCRIPTION.get('class',       None)
                    pvers  = plugin.PLUGIN_DESCRIPTION.get('version',     None)
                    pdescr = plugin.PLUGIN_DESCRIPTION.get('description', None)

                    # make sure details are complete
                    if not ptype:
                        log.error('no plugin type in %s' % pshort)
                        continue

                    if not pname:
                        log.error('no plugin name in %s' % pshort)
                        continue

                    if not pclass:
                        log.error('no plugin class in %s' % pshort)
                        continue

                    if not pvers:
                        log.error('no plugin version in %s' % pshort)
                        continue

                    if not pdescr:
                        log.error('no plugin description in %s' % pshort)
                        continue

                    # now put the plugin and plugin info into the plugin
                    # registry.  Duh!

                    pinfo = {
                        'plugin'     : plugin,
                        'class'      : pclass,
                        'type'       : ptype,
                        'name'       : pname,
                        'version'    : pvers,
                        'description': pdescr,
                        'instance'   : None
                    }
                    self.register(ptype, pname, pinfo)

                    log.debug('loading plugin %s' % pfile)
                    log.info ('loading plugin %s' % pshort)

                except Exception:
                    log.exception('loading plugin %s failed' % pshort)


    # --------------------------------------------------------------------------
    #
    def list_types(self):
        '''
        return a list of loaded plugin types
        '''
        return list(self._registry.keys())


    # --------------------------------------------------------------------------
    #
    def list(self, ptype):
        '''
        return a list of loaded plugins for a given plugin type
        '''
        if ptype not in self._registry:
            self._log.debug(self.dump_str())
            raise LookupError('No such plugin type %s in %s'
                    % (ptype, list(self._registry.keys())))

        return list(self._registry[ptype].keys())


    # --------------------------------------------------------------------------
    #
    def describe(self, ptype, pname):
        '''
        return a plugin details for a given plugin type / name pair
        '''
        if ptype not in self._registry:
            self._log.debug(self.dump_str())
            raise LookupError('No such plugin type %s in %s'
                    % (ptype, list(self._registry.keys())))

        if pname not in self._registry[ptype]:
            self._log.debug(self.dump_str())
            raise LookupError('No such plugin name %s (type: %s) in %s'
                    % (pname, ptype, list(self._registry[ptype].keys())))

        return self._registry[ptype][pname]


    # --------------------------------------------------------------------------
    #
    def load(self, ptype, pname, *args, **kwargs):
        '''
        check if a plugin with given type and name was loaded, if so,
        instantiate its plugin class and return it.
        '''

        if ptype not in self._registry:
            self._log.debug(self.dump_str())
            raise LookupError('No such plugin type %s in %s'
                    % (ptype, list(self._registry.keys())))

        if pname not in self._registry[ptype]:
            self._log.debug(self.dump_str())
            raise LookupError('No such plugin name %s (type: %s) in %s'
                    % (pname, ptype, list(self._registry[ptype].keys())))

        try:
            pdescr = self._registry[ptype][pname]
            plugin = pdescr['plugin']
            pclass = pdescr['class']
            pinst  = getattr(plugin, pclass)(pdescr, *args, **kwargs)

            assert isinstance(pinst, PluginBase), pinst

        except Exception as e:
            self._log.exception('plugin init failed')
            raise LookupError('Failed to load plugin %s (type: %s)' %
                              (pname, ptype)) from e

        return pinst


    # --------------------------------------------------------------------------
    #
    def dump(self):

        print('plugins')
        pprint.pprint(self._registry)


    # --------------------------------------------------------------------------
    #
    def dump_str(self):

        return 'plugins: \n%s' % pprint.pformat(self._registry)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2018, RADICAL"
__license__   = "GPL"


'''
Provide an abstraction for distributed queue network.

The network nodes are instances of queue_network::Node.  A node can exist either
as thread, as a local process, or as a remote process, the latter spawned via
SAGA.

The communication or work items between nodes is always via queues, were groups
of nodes feed into the same queues, and/or are fed from the same queues.  A node
can be fed from multiple queues, where queues are prioritized according to some
policy.  Depending if two communicating nodes are co-threads, co-processes, or
remote processes, different queue implementations (with the same semantic) are
used.

A network is created according to a specific state model.  That model describes
the state of entities passing through the queue network.  Each active state is
enacted by a network node, and each pending state is represented by the entity
being passed through a queue.  A queue connection thus exists for every valid
state transition.

Nodes also support the observer patter, ie. they can notify observers about
enacted state transitions.
'''

# ------------------------------------------------------------------------------
#
# http://stackoverflow.com/questions/9539052/python-dynamically-changing-base-classes-at-runtime-how-to
#
# Depending on agent architecture (which is specific to the resource type it
# runs on) can switch between different component types: using threaded (when
# running on the same node), multiprocessing (also for running on the same node,
# but avoiding python's threading problems, for the prices of slower queues),
# and remote processes (for running components on different nodes, using zeromq
# queues for communication).
#
# We do some trickery to keep the actual components independent from the actual
# schema:
#
#   - wrap the different queue types into a rpu.Queue object
#   - change the base class of the component dynamically to the respective type
#
# This requires components to adhere to the following restrictions:
#
#   - *only* communicate over queues -- no shared data with other components or
#     component instances.  Note that this also holds for example for the
#     scheduler!
#   - no shared data between the component class and it's run() method.  That
#     includes no sharing of queues.
#   - components inherit from base_component, and the constructor needs to
#     register all required component-internal and -external queues with that
#     base class -- the run() method can then transparently retrieve them from
#     there.
#


# ------------------------------------------------------------------------------
#
class Network(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, state_model=None, network_description=None):
        '''
        span the network according to state model and network description
        '''

        if state_model         is None: state_model = dict()
        if network_description is None: network_description = dict()

        self._state_model         = state_model
        self._network_description = network_description

        self._callbacks = dict()


    # --------------------------------------------------------------------------
    #
    def feed(self, entities):
        '''
        feed entities into the network
        '''

        if not isinstance(entities, list):
            entities = [entities]

        for entity in entities:
            assert entity.state == self._state_model['initial']

        self.feeder_queue.push(entities)


    # --------------------------------------------------------------------------
    #
    def subscribe(self, states, callback):
        '''
        subscribe for callback notifications for state transitions *into* the
        specified states.  If callbacks return 'False' or 'None', they are
        unregistered.
        '''

        if not isinstance(states, list):
            states = [states]

        for state in states:

            if state not in self._calbacks:
                self._callbacks[state] = list()

            self._callbacks[state].append(callback)


# ------------------------------------------------------------------------------





import os

import threading as mt

from typing import Optional

from ..logger  import Logger
from ..profile import Profiler
from ..config  import Config
from ..json_io import read_json, write_json

from .utils import LOG_ENABLED

QUEUE   = 'QUEUE'
PUBSUB  = 'PUBSUB'
UNKNOWN = 'UNKNOWN'


# ------------------------------------------------------------------------------
#
class Bridge(object):
    '''

    A bridge can be configured to have a finite lifetime: when no messages are
    received in `timeout` seconds, the bridge process will terminate.
    '''

    # --------------------------------------------------------------------------
    #
    @staticmethod
    def get_config(name, pwd=None):

        if not pwd:
            pwd = '.'

        fname = '%s/%s.cfg' % (pwd, name)

        cfg = dict()
        if os.path.isfile(fname):
            cfg = read_json(fname)

        return Config(from_dict=cfg)


    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, log=None):

        self._cfg     = cfg
        self._log     = log
        self._channel = self._cfg.channel
        self._uid     = self._cfg.uid
        self._pwd     = self._cfg.path

        if not self._pwd:
            self._pwd = os.getcwd()

        if not self._log:
            if LOG_ENABLED: level = self._cfg.log_lvl
            else          : level = 'ERROR'
            self._log = Logger(name=self._uid, ns='radical.utils.zmq',
                               level=level, path=self._pwd)

        self._prof = Profiler(name=self._uid, path=self._pwd)

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()
        else:
            self._prof.disable()

        self._prof.prof('init', uid=self._uid, msg=self._pwd)
        self._log.debug('bridge %s init', self._uid)

        self._bridge_initialize()


    # --------------------------------------------------------------------------
    #
    def write_config(self, fname=None):

        if not fname:
            fname = '%s/%s.cfg' % (self._pwd, self._cfg.uid)

        write_json(fname, {'uid'        : self._cfg.uid,
                           self.type_in : str(self.addr_in),
                           self.type_out: str(self.addr_out)})


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel

    # protocol independent addr query
    @property
    def type_in(self):
        raise NotImplementedError()

    @property
    def type_out(self):
        raise NotImplementedError()

    @property
    def addr_in(self):
        raise NotImplementedError()

    @property
    def addr_out(self):
        raise NotImplementedError()

    def _bridge_initialize(self):
        raise NotImplementedError()

    def _bridge_work(self):
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    #
    def start(self):

        # the bridge runs in a thread.  It is the bridge's owner process'
        # responsibility to ensure the thread is seeing suffient time to perform
        # as needed.  Given Python's thread performance (or lack thereof), this
        # basically means that the user of this class should create a separate
        # process instance to host the bridge thread.
        self._term          = mt.Event()
        self._bridge_thread = mt.Thread(target=self._bridge_work)
        self._bridge_thread.daemon = True
        self._bridge_thread.start()

        self._log.info('started bridge %s', self._uid)


    # --------------------------------------------------------------------------
    #
    def wait(self, timeout=None):
        '''
        wait for the bridge to terminate.  If `timeout` is set, the call will
        return after that many seconds, with a return value indicating whether
        the bridge is still alive.
        '''

        self._bridge_thread.join(timeout=timeout)

        if timeout is not None:
            return not self._bridge_thread.is_alive()


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def create(channel : str,
               kind    : Optional[str]  = None,
               cfg     : Optional[dict] = None):

        # FIXME: add other config parameters: batch size, log level, etc.

        # NOTE: I'd rather have this as class data than as stack data, but
        #       python stumbles over circular imports at that point :/
        #       Another option though is to discover and dynamically load
        #       components.

        from .pubsub import PubSub
        from .queue  import Queue

        _btypemap = {PUBSUB: PubSub,
                     QUEUE : Queue}

        if not kind:
            if   'queue'  in channel.lower(): kind = QUEUE
            elif 'pubsub' in channel.lower(): kind = PUBSUB
            else                            : kind = UNKNOWN

        if kind not in _btypemap:
            raise ValueError('unknown bridge type (%s)' % kind)

        btype  = _btypemap[kind]
        bridge = btype(channel, cfg=cfg)

        return bridge


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._term.set()
      # self._bridge_thread.join(timeout=timeout)
        self._prof.prof('term', uid=self._uid)

      # if timeout is not None:
      #     return not self._bridge_thread.is_alive()


    # --------------------------------------------------------------------------
    #
    @property
    def alive(self):
        return self._bridge_thread.is_alive()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import time

import threading     as mt

import radical.utils as ru

CHANNEL = 'test'
TOPIC   = 'test'


# ------------------------------------------------------------------------------
#
def put(channel, uid, url, n):

    put = ru.zmq.Putter(channel, url)
    for i in range(n):
        put.put(ru.as_bytes('%s: message %d' % (uid, i)))
        print('%s: message %d' % (uid, i))
        time.sleep(0.1)

    put.put(ru.as_bytes('%s: STOP' % uid))

    print('%s: done' % uid)


# ------------------------------------------------------------------------------
#
def get(channel, uid, url):

    cont = True
    get  = ru.zmq.Getter(channel, url)

    while cont:

        msgs = get.get()
        if not msgs:
            continue

        print('%s: %s' % (uid, ru.as_string(msgs)))

        for msg in msgs:
            if 'STOP' in ru.as_string(msg):
                cont = False

    print('%s: done' % uid)


# ------------------------------------------------------------------------------
#
def main():

    bridge = ru.zmq.Queue(CHANNEL)
    bridge.start()

    threads = list()

    # start some putters and getters
    # NOTE: start more putters than getters for clean termination
    for i in range(4):
        t = mt.Thread(target=put, args=(CHANNEL, 'put.%d' % i, bridge.addr_put, 5))
        t.daemon = True
        t.start()
        threads.append(t)

    for i in range(3):
        t = mt.Thread(target=get, args=(CHANNEL, 'get.%d' % i, bridge.addr_get))
        t.daemon = True
        t.start()
        threads.append(t)

    # wait for completion
    for t in threads:
        t.join()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    main()


# ------------------------------------------------------------------------------





import atexit
import shelve

from typing import List, Optional, Any

from ..json_io    import write_json
from ..dict_mixin import DictMixin

from .server import Server
from .client import Client

_registries = list()


# ------------------------------------------------------------------------------
#
def _flush_registries():
    for _reg in _registries:
        _reg.stop()


atexit.register(_flush_registries)


# ------------------------------------------------------------------------------
#
class Registry(Server):
    '''
    The `ru.zmq.Registry` is a ZMQ service which provides a hierarchical
    persistent data store.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, url       : Optional[str] = None,
                       uid       : Optional[str] = None,
                       path      : Optional[str] = None,
                       persistent: bool          = False) -> None:

        super().__init__(url=url, uid=uid, path=path)

        if persistent:
            path = '%s/%s.db' % (self._path, self._uid)
            self._log.debug('use shelve %s', path)

            self._data = shelve.open(path, writeback=True)

        else:
            self._log.debug('use in-memory dict')
            self._data = dict()

        self.register_request('put',  self.put)
        self.register_request('get',  self.get)
        self.register_request('keys', self.keys)
        self.register_request('del',  self.delitem)
        self.register_request('dump', self.dump)


    # --------------------------------------------------------------------------
    #
    def dump(self, name: str = None) -> None:


        if not isinstance(self._data, dict):
            self._log.debug('ignore dump for non-dict %s', name)

        else:
            if name: fname = '%s/%s.%s.json' % (self._path, self._uid, name)
            else   : fname = '%s/%s.json'    % (self._path, self._uid)

            self._log.debug('dumo to %s', fname)
            write_json(self._data, fname)



    # --------------------------------------------------------------------------
    #
    def stop(self) -> None:

      # self.dump()

        self._log.debug('stop')

        if isinstance(self._data, shelve.Shelf):
            self._data.close()

        super().stop()


    # --------------------------------------------------------------------------
    #
    def put(self, key: str, val: Any) -> None:

        this  = self._data
        elems = key.split('.')
        path  = elems[:-1]
        leaf  = elems[-1]

        self._log.debug_9('put %s: %s', str(key), str(val))

        for elem in path:

            if elem not in this or this[elem] is None:
                this[elem] = dict()

            this = this[elem]

        this[leaf] = val

        if not isinstance(self._data, dict):
            self._data.sync()


    # --------------------------------------------------------------------------
    #
    def get(self, key: str) -> Optional[str]:

        this  = self._data
        elems = key.split('.')
        path  = elems[:-1]
        leaf  = elems[-1]

        for elem in path:
            this = this.get(elem, {})
            if not this:
                break

        if this is None:
            this = dict()

        val = this.get(leaf)

        self._log.debug_9('get %s: %s', str(key), str(val))
        return val


    # --------------------------------------------------------------------------
    #
    def keys(self, pwd: Optional[str] = None) -> List[str]:

        this = self._data

        if pwd:
            path = pwd.split('.')
            for elem in path:
                this = this.get(elem, {})
                if not this:
                    break

        if this is None:
            this = dict()

        keys = list(this.keys())

        self._log.debug_9('keys: %s', keys)

        return keys


    # --------------------------------------------------------------------------
    #
    def delitem(self, key: str) -> None:

        self._log.debug_9('del: %s', key)

        this = self._data

        if key:
            path = key.split('.')
            for elem in path[:-1]:
                this = this.get(elem, {})
                if not this:
                    break

            if this:
                del this[path[-1]]


# ------------------------------------------------------------------------------
#
class RegistryClient(Client, DictMixin):
    '''
    The `ru.zmq.RegistryClient` class provides a simple dict-like interface to
    a remote `ru.zmq.Registry` service.  Note that only top-level dict-actions
    on the `RegistryClient` instance are synced with the remote service storage.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, url: str,
                       pwd: Optional[str] = None) -> None:

        self._url = url
        self._pwd = pwd

        super().__init__(url=url)


    # --------------------------------------------------------------------------
    #
    def dump(self, name: str = None) -> None:

        return self.request(cmd='dump', name=name)


    # --------------------------------------------------------------------------
    # verbose API
    def get(self, key    : str,
                  default: Optional[Any] = None) -> Optional[Any]:

        if self._pwd:
            key = self._pwd + '.' + key

        try:
            return self.request(cmd='get', key=key)
        except:
            return default


    def put(self, key: str,
                  val: Any) -> None:

        if self._pwd:
            key = self._pwd + '.' + key
        ret = self.request(cmd='put', key=key, val=val)

        assert ret is None
        return ret


    # --------------------------------------------------------------------------
    # dict mixin API
    def __getitem__(self, key: str) -> Optional[Any]:

        return self.get(key)


    def __setitem__(self, key: str, val: Any) -> None:

        return self.put(key, val)


    def __delitem__(self, key: str) -> None:

        if self._pwd:
            key = self._pwd + '.' + key
        ret = self.request(cmd='del', key=key)
        assert ret is None


    def keys(self) -> List[str]:

        ret = self.request(cmd='keys', pwd=self._pwd)
        assert isinstance(ret, list)
        return ret


# ------------------------------------------------------------------------------





import os
import queue
import shlex
import select

import threading  as mt
import subprocess as sp

from .constants import RUNNING, DONE, FAILED
from .misc      import is_string
from .misc      import ru_open


# ------------------------------------------------------------------------------
#
def sh_quote(data):
    '''
    quote a string and wrap it in double quotes so that it can passed to a POSIX
    shell.

    Examples:

      foo        -> "foo"
      foo"bar    -> "foo\\"bar"
      foo\\"bar  -> "foo\\\\\\"bar"
      $FOO'$BAR  -> "$FOO'$BAR"

    NOTE: this method does not attempt to strip out backticks or other code
          execution mechanisms from the string.

    '''

    if not data:
        return '""'

    data = data.replace('\\', '\\\\')
    data = data.replace('"',  '\\"')

    return '"%s"' % data


# ------------------------------------------------------------------------------
#
def sh_callout(cmd, stdout=True, stderr=True,
                    shell=False, env=None, cwd=None):
    '''
    call a shell command, return `[stdout, stderr, retval]`.
    '''

    # convert string into arg list if needed
    if is_string(cmd) and \
       not shell: cmd    = shlex.split(cmd)

    if stdout   : stdout = sp.PIPE
    else        : stdout = None

    if stderr   : stderr = sp.PIPE
    else        : stderr = None

    p = sp.Popen(cmd, stdout=stdout, stderr=stderr,
                      shell=shell, env=env, cwd=cwd)

    if not stdout and not stderr:
        ret = p.wait()
    else:
        stdout, stderr = p.communicate()
        ret            = p.returncode

    return stdout.decode("utf-8"), stderr.decode("utf-8"), ret


# ------------------------------------------------------------------------------
#
def sh_callout_bg(cmd, stdout=None, stderr=None,
                       shell=False, env=None, cwd=None):
    '''
    call a shell command in the background.  Do not attempt to pipe STDOUT/ERR,
    but only support writing to named files.
    '''

    # pipes won't work - see sh_callout_async
    if stdout == sp.PIPE: raise ValueError('stdout pipe unsupported')
    if stderr == sp.PIPE: raise ValueError('stderr pipe unsupported')

    # openfile descriptors for I/O, if needed
    if is_string(stdout): stdout = ru_open(stdout, 'w')
    if is_string(stderr): stderr = ru_open(stderr, 'w')

    # convert string into arg list if needed
    if not shell and is_string(cmd): cmd = shlex.split(cmd)

    sp.Popen(cmd, stdout=stdout, stderr=stderr, shell=shell, env=env, cwd=cwd)

    return


# ------------------------------------------------------------------------------
#
def sh_callout_async(cmd, stdin=True, stdout=True, stderr=True,
                          shell=False, env=None, cwd=None):
    '''

    Run a command, and capture stdout/stderr if so flagged.  The call will
    return an PROC object instance on which the captured output can be retrieved
    line by line (I/O is line buffered).  When the process is done, a `None`
    will be returned on the I/O queues.

    Line breaks are stripped.

    stdout/stderr: True [default], False, string
      - False : discard I/O
      - True  : capture I/O as queue [default]
      - string: capture I/O as queue, also write to named file

    shell: True, False [default]
      - pass to popen

    cwd: string
      - working directory for command to run in

    PROC:
      - PROC.stdout         : `queue.Queue` instance delivering stdout lines
      - PROC.stderr         : `queue.Queue` instance delivering stderr lines
      - PROC.state          : ru.RUNNING, ru.DONE, ru.FAILED
      - PROC.rc             : returncode (None while ru.RUNNING)
      - PROC.stdout_filename: name of stdout file (when available)
      - PROC.stderr_filename: name of stderr file (when available)
    '''
    # NOTE: Fucking python screws up stdio buffering when threads are used,
    #       *even if the treads do not perform stdio*.  Its possible that the
    #       logging module interfers, too.  Either way, I am fed up debugging
    #       this shit, and give up.  This method does not work for threaded
    #       python applications.
    assert False, 'this is broken for python apps'

    # --------------------------------------------------------------------------
    #
    class _P(object):
        '''
        internal representation of a process
        '''

        # ----------------------------------------------------------------------
        def __init__(self, cmd, stdin, stdout, stderr, shell, env, cwd):

            cmd = cmd.strip()

            self._in_c  = bool(stdin)                # flag stdin  capture
            self._out_c = bool(stdout)               # flag stdout capture
            self._err_c = bool(stderr)               # flag stderr capture

            self._in_r , self._in_w  = os.pipe()     # put stdin  to   child
            self._out_r, self._out_w = os.pipe()     # get stdout from child
            self._err_r, self._err_w = os.pipe()     # get stderr from child

            self._in_o  = os.fdopen(self._in_r)      # file object for in  ep
            self._out_o = os.fdopen(self._out_r)     # file object for out ep
            self._err_o = os.fdopen(self._err_r)     # file object for err ep

            self._in_q  = queue.Queue()              # get stdin  from parent
            self._out_q = queue.Queue()              # put stdout to   parent
            self._err_q = queue.Queue()              # put stderr to   parent

            if is_string(stdout): self._out_f = ru_open(stdout, 'w')
            else                : self._out_f = None

            if is_string(stderr): self._err_f = ru_open(stderr, 'w')
            else                : self._err_f = None

            self.state = RUNNING
            self._proc = sp.Popen(cmd, stdin=self._in_r,
                                       stdout=self._out_w,
                                       stderr=self._err_w,
                                       shell=shell,
                                       env=env,
                                       cwd=cwd,
                                       bufsize=1)

            t = mt.Thread(target=self._watch)
            t.daemon = True
            t.start()

            self.rc = None  # return code


        @property
        def stdin(self):
            if not self._in_c:
                raise RuntimeError('stdin not captured')
            return self._in_q


        @property
        def stdout(self):
            if not self._out_c:
                raise RuntimeError('stdout not captured')
            return self._out_q


        @property
        def stderr(self):
            if not self._err_c:
                raise RuntimeError('stderr not captured')
            return self._err_q


        @property
        def stdout_filename(self):
            if not self._out_f:
                raise RuntimeError('stdout not recorded')
            return self._out_f.name


        @property
        def stderr_filename(self):
            if not self._err_f:
                raise RuntimeError('stderr not recorded')
            return self._err_f.name


        def kill(self):
            self._proc.terminate()


        # ----------------------------------------------------------------------
        #
        def _watch(self):

            poller = select.poll()
            poller.register(self._out_r, select.POLLIN | select.POLLHUP)
            poller.register(self._err_r, select.POLLIN | select.POLLHUP)

            # try forever to read stdin, stdout and stderr, stop only when
            # either signals that process (parent or child) died
            while True:

                # check for input
                data = self._in_q.get_nowait()
                if data:
                    self._out_o.write(data)
                    self._out_f.write(data)

                active = False
                fds    = poller.poll(100)  # timeout configurable (ms)

                for fd,mode in fds:

                    if mode & select.POLLHUP:
                        # fd died - grab data from other fds
                        continue

                    if fd    == self._out_r:
                        o_in  = self._out_o
                        q_out = self._out_q
                        f_out = self._out_f

                    elif fd  == self._err_r:
                        o_in  = self._err_o
                        q_out = self._err_q
                        f_out = self._err_f

                    line = o_in.readline()  # `bufsize=1` in `popen`

                    if line:
                        # found valid data (active)
                        active = True
                        if q_out: q_out.put(line.rstrip('\n'))
                        if f_out: f_out.write(line)

                # no data received - check process health
                if not active and self._proc.poll() is not None:

                    # process is dead
                    self.rc = self._proc.returncode

                    if self.rc == 0: self.state = DONE
                    else           : self.state = FAILED

                    if self._out_q: self._out_q.put(None)  # signal EOF
                    if self._err_q: self._err_q.put(None)  # signal EOF

                    if self._out_q: self._out_q.join()     # ensure reads
                    if self._err_q: self._err_q.join()     # ensure reads

                    return  # finishes thread
    # --------------------------------------------------------------------------

    return _P(cmd=cmd, stdin=stdin, stdout=stdout, stderr=stderr,
                       shell=shell, env=env, cwd=cwd)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2018, RADICAL"
__license__   = "GPL"


"""
Provide an abstract base class for a component of a distributed queue network.
The node can exist either as thread, as a local process, or as a remote process,
spawned via SAGA.  The communication between nodes is always via queues, were
groups of nodes feed into the same queues, and/or are fed from the same queues.
A node can be fed from multiple queues, where queues are prioritized according
to some policy.
"""


# ------------------------------------------------------------------------------
#
class Node(object):
    pass


# ------------------------------------------------------------------------------





# flake8: noqa: F401
#
__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2013, RADICAL"
__license__   = "GPL"

# this is work in progress and should not be used
# pylint: disable=import-error

from .node    import Node
from .queue   import Queue
from .network import Network




#!/usr/bin/env python

import zmq
import time
import msgpack

# ------------------------------------------------------------------------------
#
delay = 0.0

# src side is proper push/pull (queue)
context       = zmq.Context()
socket_in     = context.socket(zmq.PULL)
socket_in.hwm = 1
socket_in.bind("tcp://*:*")

# out side is req/resp for load balancing
context        = zmq.Context()
socket_out     = context.socket(zmq.REP)
socket_out.hwm = 1
socket_out.bind("tcp://*:*")

addr_in  = socket_in .getsockopt(zmq.LAST_ENDPOINT)
addr_out = socket_out.getsockopt(zmq.LAST_ENDPOINT)

print('PUT: %s' % addr_in)
print('GET: %s' % addr_out)

with open('test.bridge.url', 'w') as fout:
    fout.write('PUT %s\n' % addr_in)
    fout.write('GET %s\n' % addr_out)

# zmq.proxy(socket_in, socket_out)

while True:
    # only read from socket_in once we have a consumer requesting the data on
    # socket_out.
    # TODO: For multiple sinks, make sure we send replies to the right one
    req = socket_out.recv()
    msg = msgpack.unpackb(socket_in.recv())
    msg['req'] = req
    print('<> %s' % msg)
    socket_out.send(msgpack.packb(msg))

    time.sleep(delay)


# ------------------------------------------------------------------------------




# pylint: disable=protected-access, abstract-class-instantiated

import sys
import zmq
import time

import threading as mt

from typing      import Optional

from ..atfork    import atfork
from ..config    import Config
from ..ids       import generate_id, ID_CUSTOM
from ..misc      import as_string, as_bytes, as_list, noop
from ..logger    import Logger
from ..profile   import Profiler
from ..debug     import print_exception_trace
from ..serialize import to_msgpack, from_msgpack

from .bridge     import Bridge
from .utils      import zmq_bind, no_intr
from .utils      import log_bulk, LOG_ENABLED
# from .utils    import prof_bulk


# --------------------------------------------------------------------------
#
_LINGER_TIMEOUT    =  250  # ms to linger after close
_HIGH_WATER_MARK   =    0  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE = 1024  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
def _atfork_child():
    Getter._callbacks = dict()                                            # noqa


atfork(noop, noop, _atfork_child)


# ------------------------------------------------------------------------------
#
# Communication between components is done via queues.  Queues are
# uni-directional, ie. Queues have an input-end for which one can call 'put()',
# and and output-end, for which one can call 'get()'.
#
# The semantics we expect (and which is what is matched by the native Python
# `Queue.Queue`), is:
#
#   - multiple upstream   components put messages onto the same queue (input)
#   - multiple downstream components get messages from the same queue (output)
#   - local order of messages is maintained: order of messages pushed onto the
#     *same* input is preserved when pulled on any output
#   - message routing is fair: whatever downstream component calls 'get' first
#     will get the next message
#
# We implement the interface of Queue.Queue:
#
#   put(msg)
#   get()
#   get_nowait()
#
# Not implemented is, at the moment:
#
#   qsize
#   empty
#   full
#   put(msg, block, timeout)
#   put_nowait
#   get(block, timeout)
#   task_done
#
# Our Queue additionally takes 'name', 'role' and 'address' parameter on the
# constructor.  'role' can be 'input', 'bridge' or 'output', where 'input' is
# the end of a queue one can 'put()' messages into, and 'output' the end of the
# queue where one can 'get()' messages from. A 'bridge' acts as as a message
# forwarder.  'address' denominates a connection endpoint, and 'name' is
# a unique identifier: if multiple instances in the current process space use
# the same identifier, they will get the same queue instance (are connected to
# the same bridge).
#
class Queue(Bridge):

    def __init__(self, channel: str, cfg: Optional[dict] = None, log=None):
        '''
        This Queue type sets up an zmq channel of this kind:

            input \\              // output
                     == bridge ==
            input //              \\ output

        ie. any number of inputs can 'zmq.push()' to a bridge (which
        'zmq.pull()'s), and any number of outputs can 'zmq.request()'
        messages from the bridge (which 'zmq.response()'s).

        The bridge is the entity which 'bind()'s network interfaces, both input
        and output type endpoints 'connect()' to it.  It is the callees
        responsibility to ensure that only one bridge of a given type exists.

        Addresses are of the form 'tcp://host:port'.  Both 'host' and 'port' can
        be wildcards for BRIDGE roles -- the bridge will report the in and out
        addresses as obj.addr_put and obj.addr_get.
        '''

        if cfg:
            # create deep copy
            cfg = Config(cfg=cfg)
        else:
            cfg = Config()

        # ensure channel is set in config
        if cfg.channel:
            assert cfg.channel == channel
        else:
            cfg.channel = channel

        if not cfg.uid:
            cfg.uid = generate_id('%s.bridge.%%(counter)04d' % cfg.channel,
                                  ID_CUSTOM)

        super().__init__(cfg, log=log)

        self._bulk_size  = self._cfg.get('bulk_size', 0)

        if self._bulk_size <= 0:
            self._bulk_size = _DEFAULT_BULK_SIZE


    # --------------------------------------------------------------------------
    #

    # protocol independent addr query
    @property
    def type_in(self):
        return 'put'

    @property
    def type_out(self):
        return 'get'

    @property
    def addr_in(self):
        return self._addr_put

    @property
    def addr_out(self):
        return self._addr_get

    # protocol dependent addr query
    @property
    def addr_put(self):
        return self._addr_put

    @property
    def addr_get(self):
        return self._addr_get


    # --------------------------------------------------------------------------
    #
    def _bridge_initialize(self):

        self._log.info('start bridge %s', self._uid)

        self._lock       = mt.Lock()

        self._ctx        = zmq.Context()  # rely on GC for destruction
        self._put        = self._ctx.socket(zmq.PULL)
        self._put.linger = _LINGER_TIMEOUT
        self._put.hwm    = _HIGH_WATER_MARK
        self._addr_put   = zmq_bind(self._put)

        self._get        = self._ctx.socket(zmq.REP)
        self._get.linger = _LINGER_TIMEOUT
        self._get.hwm    = _HIGH_WATER_MARK
        self._addr_get   = zmq_bind(self._get)

        self._log.info('bridge in  %s: %s', self._uid, self._addr_put)
        self._log.info('bridge out %s: %s', self._uid, self._addr_get)

        # start polling senders
        self._poll_put = zmq.Poller()
        self._poll_put.register(self._put, zmq.POLLIN)

        # start polling receivers
        self._poll_get = zmq.Poller()
        self._poll_get.register(self._get, zmq.POLLIN)


    # --------------------------------------------------------------------------
    #
    def _bridge_work(self):

        # TODO: *always* pull for messages and buffer them.  Serve requests from
        #       that buffer.

        try:

            self.nin  = 0
            self.nout = 0
            self.last = 0

            buf = dict()
            while not self._term.is_set():

                active = False

                # check for incoming messages, and buffer them
                ev_put = dict(no_intr(self._poll_put.poll, timeout=10))
              # self._prof.prof('poll_put', msg=len(ev_put))
                self._log.debug_9('polled put: %s', ev_put)

                if self._put in ev_put:

                    with self._lock:
                        data = list(no_intr(self._put.recv_multipart))
                    self._log.debug_9('recvd  put: %s', data)

                    if len(data) != 2:
                        raise RuntimeError('%d frames unsupported' % len(data))

                    qname = as_string(from_msgpack(data[0]))
                    msgs  = from_msgpack(data[1])
                  # prof_bulk(self._prof, 'poll_put_recv', msgs)
                    log_bulk(self._log, '<> %s' % qname, msgs)
                    self._log.debug_9('put %s: %s ! ', qname, len(msgs))

                    if qname not in buf:
                        buf[qname] = list()
                    buf[qname] += msgs
                    self.nin   += len(msgs)

                    active = True


                # check if somebody wants our messages
                ev_get = dict(no_intr(self._poll_get.poll, timeout=10))
              # self._prof.prof('poll_get', msg=len(ev_get))
                self._log.debug_9('polled get: %s [%s]', ev_get, self._get)

                if self._get in ev_get:

                    # send up to `bulk_size` messages from the buffer
                    # NOTE: this sends partial bulks on buffer underrun
                    with self._lock:
                        # the actual req message is ignored - we only care
                        # about who sent it
                        qname = as_string(no_intr(self._get.recv))

                    if not qname:
                        qname = 'default'

                    if qname in buf:
                        msgs = buf[qname][:self._bulk_size]
                    else:
                        self._log.debug_9('get: %s not in %s', qname,
                                                             list(buf.keys()))
                        msgs = list()

                    log_bulk(self._log, '>< %s' % qname, msgs)

                    data   = [to_msgpack(qname), to_msgpack(msgs)]
                    active = True

                  # self._log.debug_9('==== get %s: %s', qname, list(buf.keys()))
                  # self._log.debug_9('==== get %s: %s', qname, list(buf.values()))
                  # self._log.debug_9('==== get %s: %s ! [%s]', qname, len(msgs),
                  #                        [[x, len(y)] for x,y in buf.items()])
                    no_intr(self._get.send_multipart, data)
                  # prof_bulk(self._prof, 'poll_get_send', msgs=msgs, msg=req)

                    self.nout += len(msgs)
                    self.last  = time.time()

                    # remove sent messages from buffer
                    if msgs:
                        del buf[qname][:self._bulk_size]

                if not active:
                  # self._prof.prof('sleep', msg=len(buf))
                    # let CPU sleep a bit when there is nothing to do
                    # We don't want to use poll timouts since we use two
                    # competing polls and don't want the idle channel slow down
                    # the busy one.
                    time.sleep(0.01)

        except Exception:
            self._log.exception('bridge failed')

    def stop(self):
        Bridge.stop(self)


# ------------------------------------------------------------------------------
#
class Putter(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, log=None, prof=None, path=None):

        self._channel  = channel
        self._url      = as_string(url)
        self._log      = log
        self._prof     = prof
        self._lock     = mt.Lock()
        self._uid      = generate_id('%s.put.%%(counter)04d' % self._channel,
                                     ID_CUSTOM)

        if not self._url:
            self._url = Bridge.get_config(channel, path).get('put')

        if not self._url:
            raise ValueError('no contact url specified, no config found')

        if not self._log:
            if LOG_ENABLED: level = 'DEBUG_9'
            else          : level = 'ERROR'
            self._log = Logger(name=self._uid, ns='radical.utils.zmq',
                               level=level, path=path)

        if not self._prof:
            self._prof = Profiler(name=self._uid, ns='radical.utils', path=path)
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect put to %s: %s', self._channel, self._url)

        self._ctx      = zmq.Context()  # rely on GC for destruction
        self._q        = self._ctx.socket(zmq.PUSH)
        self._q.linger = _LINGER_TIMEOUT
        self._q.hwm    = _HIGH_WATER_MARK
        self._q.connect(self._url)


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        return 'Putter(%s @ %s)'  % (self.channel, self._url)

    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def put(self, msgs, qname=None):

        msgs = as_list(msgs)

        if not qname:
            qname = 'default'

        log_bulk(self._log, '-> %s[%s]' % (self._channel, qname), msgs)
        data = [to_msgpack(qname), to_msgpack(msgs)]

        with self._lock:
            no_intr(self._q.send_multipart, data)
      # prof_bulk(self._prof, 'put', msgs)


# ------------------------------------------------------------------------------
#
class Getter(object):

    # instead of creating a new listener thread for each endpoint which then, on
    # incoming messages, calls a getter callback, we only create *one*
    # listening thread per ZMQ endpoint address and call *all* registered
    # callbacks in that thread.  We hold those endpoints in a class dict, so
    # that all class instances share that information
    _callbacks = dict()


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_nowait(url, qname=None, timeout=None, uid=None):  # timeout in ms

        info = Getter._callbacks[url]

        if not qname:
            qname = 'default'

        with info['lock']:

            if LOG_ENABLED: level = 'DEBUG_9'
            else          : level = 'ERROR'
            logger = Logger(name=qname, ns='radical.utils.zmq', level=level)

            if not info['requested']:

                # send the request *once* per recieval (got lock above)
                # FIXME: why is this sent repeatedly?
                logger.debug_9('=> from %s[%s]', uid, qname)
                no_intr(info['socket'].send, as_bytes(qname))
                info['requested'] = True


            if no_intr(info['socket'].poll, flags=zmq.POLLIN, timeout=timeout):

                data = list(no_intr(info['socket'].recv_multipart))
                info['requested'] = False

                qname = as_string(from_msgpack(data[0]))
                msgs  = as_string(from_msgpack(data[1]))
                log_bulk(logger, '<-1 %s [%s]' % (uid, qname), msgs)
                return msgs

            else:
                return None


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _listener(url, qname=None, uid=None):
        '''
        other than the pubsub listener, the queue listener will not deliver
        an incoming message to all subscribers, but only to exactly *one*
        subscriber.  We this perform a round-robin over all known callbacks
        '''

        if not qname:
            qname = 'default'

        assert url in Getter._callbacks
        time.sleep(0.01)

        try:
            term = Getter._callbacks.get(url, {}).get('term')
            idx  = 0  # round-robin cb index
            while not term.is_set():

                # this list is dynamic
                callbacks = Getter._callbacks[url]['callbacks']

                if not callbacks:
                    time.sleep(0.01)
                    continue

                msgs = Getter._get_nowait(url, qname=qname, timeout=500, uid=uid)
                BULK = True
                if msgs:

                    if BULK:
                        idx += 1
                        if idx >= len(callbacks):
                            idx = 0
                        cb, _lock = callbacks[idx]
                        if _lock:
                            with _lock:
                                cb(as_string(msgs))
                        else:
                            cb(as_string(msgs))

                    else:
                        for m in as_list(msgs):

                            idx += 1
                            if idx >= len(callbacks):
                                idx = 0

                            cb, _lock = callbacks[idx]
                            if _lock:
                                with _lock:
                                    cb(as_string(m))
                            else:
                                cb(as_string(m))

        except Exception as e:
            print_exception_trace()
            sys.stderr.write('listener died: %s : %s : %s\n'
                            % (qname, url, repr(e)))
            sys.stderr.flush()


    # --------------------------------------------------------------------------
    #
    def _start_listener(self, qname=None):

        if not qname:
            qname = 'default'

        # only start if needed
        if Getter._callbacks[self._url]['thread']:
            return

        t = mt.Thread(target=Getter._listener, args=[self._url, qname, self._uid])
        t.daemon = True
        t.start()

        Getter._callbacks[self._url]['thread'] = t


    # --------------------------------------------------------------------------
    #
    def _stop_listener(self, force=False):

        # only stop listener if no callbacks remain registered (unless forced)
        if force or not Getter._callbacks[self._url]['callbacks']:
            if  Getter._callbacks[self._url]['thread']:
                Getter._callbacks[self._url]['term'  ].set()
                Getter._callbacks[self._url]['thread'].join()
                Getter._callbacks[self._url]['term'  ].unset()
                Getter._callbacks[self._url]['thread'] = None


    # --------------------------------------------------------------------------
    #
    def __init__(self, channel, url=None, cb=None,
                                log=None, prof=None, path=None):
        '''
        When a callback `cb` is specified, then the Getter c'tor will spawn
        a separate thread which continues to listen on the channel, and the
        cb is invoked on any incoming message.  The message will be the only
        argument to the cb.
        '''

        self._channel   = channel
        self._url       = as_string(url)
        self._lock      = mt.Lock()
        self._log       = log
        self._prof      = prof
        self._uid       = generate_id('%s.get.%%(counter)04d' % self._channel,
                                      ID_CUSTOM)

        if not self._url:
            self._url = Bridge.get_config(channel, path).get('get')

        if not self._url:
            raise ValueError('no contact url specified, no config found')

        if not self._log:
            if LOG_ENABLED: level = 'DEBUG_9'
            else          : level = 'ERROR'
            self._log = Logger(name=self._uid, ns='radical.utils.zmq',
                               level=level, path=path)

        if not self._prof:
            self._prof  = Profiler(name=self._uid, ns='radical.utils', path=path)
            self._prof.disable()

        if 'hb' in self._uid or 'heartbeat' in self._uid:
            self._prof.disable()

        self._log.info('connect get to %s: %s', self._channel, self._url)

        self._requested = False          # send/recv sync
        self._ctx       = zmq.Context()  # rely on GC for destruction
        self._q         = self._ctx.socket(zmq.REQ)
        self._q.linger  = _LINGER_TIMEOUT
        self._q.hwm     = _HIGH_WATER_MARK
        self._q.connect(self._url)

        if url not in Getter._callbacks:

            Getter._callbacks[url] = {'uid'      : self._uid,
                                      'socket'   : self._q,
                                      'channel'  : self._channel,
                                      'lock'     : mt.Lock(),
                                      'term'     : mt.Event(),
                                      'requested': self._requested,
                                      'thread'   : None,
                                      'callbacks': list()}
        if cb:
            self.subscribe(cb)
        else:
            self._interactive = True


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        return 'Getter(%s @ %s)'  % (self.channel, self._url)

    @property
    def name(self):
        return self._uid

    @property
    def uid(self):
        return self._uid

    @property
    def channel(self):
        return self._channel


    # --------------------------------------------------------------------------
    #
    def subscribe(self, cb, lock=None):

        # if we need to serve callbacks, then open a thread to watch the socket
        # and register the callbacks.  If a thread is already runnning on that
        # channel, just register the callback.
        #
        # Note that once a thread is watching a socket, we cannot allow to use
        # `get()` and `get_nowait()` anymore, as those will interfere with the
        # thread consuming the messages,
        #
        # The given lock (if any) is used to shield concurrent cb invokations.
        #
        # FIXME: clean up lock usage - see self._lock

        if self._url not in Getter._callbacks:

            Getter._callbacks[self._url] = {'uid'      : self._uid,
                                            'socket'   : self._q,
                                            'channel'  : self._channel,
                                            'lock'     : mt.Lock(),
                                            'term'     : mt.Event(),
                                            'requested': self._requested,
                                            'thread'   : None,
                                            'callbacks': list()}

        # we allow only one cb per queue getter process at the moment, until we
        # have more clarity on the RR behavior of concurrent callbacks.
        if Getter._callbacks[self._url]['callbacks']:
            raise RuntimeError('multiple callbacks not supported')

        Getter._callbacks[self._url]['callbacks'].append([cb, lock])

        self._interactive = False
        self._start_listener()


    # --------------------------------------------------------------------------
    #
    def unsubscribe(self, cb):

        if self._url in Getter._callbacks:
            for _cb, _lock in Getter._callbacks[self._url]['callbacks']:
                if cb == _cb:
                    Getter._callbacks[self._url]['callbacks'].remove([_cb, _lock])
                    break

        self._stop_listener()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._stop_listener(force=True)


    # --------------------------------------------------------------------------
    #
    def get(self, qname=None):

        if not self._interactive:
            raise RuntimeError('invalid get(): callbacks are registered')

        if not qname:
            qname = 'default'


        # double-check: minimize lock use which is only needed for a very
        # rare race anyway
        if not self._requested:
            with self._lock:
                if not self._requested:
                    self._log.debug_9('=> from %s[%s]', self._channel, qname)
                    no_intr(self._q.send, as_bytes(qname))
                    self._requested = True

          # self._prof.prof('requested')

        with self._lock:
            data = list(no_intr(self._q.recv_multipart))
            self._requested = False

        qname = from_msgpack(data[0])
        msgs  = from_msgpack(data[1])

        log_bulk(self._log, '<-2 %s [%s]' % (self._channel, qname), msgs)

        return as_string(msgs)


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, qname=None, timeout=None):  # timeout in ms

        if not self._interactive:
            raise RuntimeError('invalid get(): callbacks are registered')

        # backward compatibility to `get_nowait(timeout=None)`
        if timeout is None and isinstance(qname, int):
            timeout = qname
            qname   = None

        if not qname:
            qname = 'default'

        if not self._requested:
            with self._lock:  # need to protect self._requested
                if not self._requested:
                    self._log.debug_9('=> from %s[%s]', self._channel, qname)
                    no_intr(self._q.send_multipart, [as_bytes(qname)])
                    self._requested = True

        if no_intr(self._q.poll, flags=zmq.POLLIN, timeout=timeout):
            with self._lock:
                data = list(no_intr(self._q.recv_multipart))
                self._requested = False

            qname = from_msgpack(data[0])
            msgs  = from_msgpack(data[1])
            log_bulk(self._log, '<-3 %s [%s]' % (self._channel, qname), msgs)

            return as_string(msgs)

        else:
            return None


# ------------------------------------------------------------------------------
#
def test_queue(channel, addr_pub, addr_sub):

    c_a  = 200
    c_b  = 400
    data = dict()

    for i in 'ABCD':
        data[i] = dict()
        for j in 'AB':
            data[i][j] = 0

    def cb(uid, msg):
        if msg['idx'] is None:
            return False
        data[uid][msg['src']] += 1

    cb_C = lambda t,m: cb('C', m)
    cb_D = lambda t,m: cb('D', m)

    Getter(channel=channel, url=addr_sub, cb=cb_C)
    Getter(channel=channel, url=addr_sub, cb=cb_D)

    # --------------------------------------------------------------------------
    def work_pub(uid, n, delay):

        pub = Putter(channel=channel, url=addr_pub)
        idx = 0

        while idx < n:
            time.sleep(delay)
            pub.put({'src': uid,
                     'idx': idx})
            idx += 1
            data[uid][uid] += 1

        # send EOF
        pub.put({'src': uid,
                 'idx': None})
    # --------------------------------------------------------------------------

    t_a = mt.Thread(target=work_pub, args=['A', c_a, 0.001])
    t_b = mt.Thread(target=work_pub, args=['B', c_b, 0.001])

    t_a.start()
    t_b.start()

    t_a.join()
    t_b.join()

    time.sleep(0.01)

    import pprint
    pprint.pprint(data)

    assert data['A']['A'] == c_a
    assert data['B']['B'] == c_b

    assert data['C']['A'] + data['C']['B'] + \
           data['D']['A'] + data['D']['B'] == 2 * (c_a + c_b)

    return data


# ------------------------------------------------------------------------------





import math as m

from .logger import Logger


# ------------------------------------------------------------------------------
#
def collapse_ranges (ranges):
    '''
    Given be a set of ranges (as a set of pairs of floats [start, end] with
    'start <= end'.  This algorithm will then collapse that set into the
    smallest possible set of ranges which cover the same, but not more nor
    less, of the domain (floats).

    We first sort the ranges by their starting point.  We then start with the
    range with the smallest starting point [start_1, end_1], and compare to
    the next following range [start_2, end_2], where we now know that
    start_1 <= start_2.  We have now two cases:

    a) when start_2 <= end_1, then the ranges overlap, and we collapse them into
    range_1: range_1 = [start_1, max[end_1, end_2]

    b) when start_2 > end_1, then ranges don't overlap.  Importantly, none of
    the other later ranges can ever overlap range_1, because there start points
    are even larger.  So we move range_1 to the set of final ranges, and restart
    the algorithm with range_2 being the smallest one.

    Termination condition is if only one range is left -- it is also moved to
    the list of final ranges then, and that list is returned.
    '''

    # FIXME: does tuple and set conversion really add anything?

    # Ranges must be unique: we do not count timings when they start and end at
    # exactly the same time. By using a set, we do not repeat ranges.
    # we convert to a list before return.
    final = set()

    # return empty list if given an empty list
    if not ranges:
        return []

    START = 0
    END   = 1

    for _range in ranges:
        if _range[START] > _range[END]:
            _range[START], _range[END] = _range[END], _range[START]

    # sort ranges into a copy list, by their start time
    _ranges = sorted(ranges, key=lambda x: x[START])

    # set 'base' to the earliest range (smallest starting point)
    base = _ranges[0]

    for _range in _ranges[1:]:

        # if range is empty, skip it
        if _range[START] == _range[END]:
            continue

        if _range[START] <= base[END]:
            # ranges overlap -- extend the base
            base[END] = max(base[END], _range[END])

        else:
            # ranges don't overlap -- move base to final, and current _range
            # becomes the new base
            final.add(tuple(base))
            base = _range

    # termination: push last base to final
    final.add(tuple(base))

    # Return final as list of list in case a mutable type is needed.
    return sorted([list(b) for b in final])


# ------------------------------------------------------------------------------
#
def range_concurrency(ranges):
    '''
    given a set of uncollapsed ranges, return a series which describes the
    range-concurrency at any point.

    Example:
      Ranges:
        [----]  [---------]             [--------]
          [--------]         [-]      [------------]      [--]

      Concurrency:
        1 2  1  2  1      0  1 0      1 2        1 0      1  0

    Returned is a sorted list of tuples where the first entry defines at what
    range value the concurrency changed, and the second value defines to what
    the concurrency count changed at that point.

    You could consider the ranges to be of type `[time_start, time_end]`, and
    the return would be a list of `[timestamp, concurrency]`, if that helps --
    but the algorithm does not make any assumption on the data type, really,
    only that the values can be sorted.
    '''

    if not ranges:
        return list()

    START = 0
    END   = 1

    for _range in ranges:
        if _range[START] > _range[END]:
            _range[START], _range[END] = _range[END], _range[START]

    # we want to sort all range boundaries by value, but also want to remember
    # if they were start or end
    times  = list()
    for idx in range(len(ranges)):
        r = ranges[idx]
        times.append([r[0], START])
        times.append([r[1], END  ])

    times.sort()

    # go through the sorted list of times, and increase concurrency on `START`,
    # decrease it on `END`.  Make sure we add up all entries for the same value
    # at once.

    # return a sequence of [timestamp, concurrency] tuples
    ret = list()

    # check initial conditions
    assert times[0][0] >= 0    , 'negative time %s'       % times[0]
    assert times[0][1] == START, 'inconsistent ranges %s' % ranges

    last_time   = times[0][0]
    concurrency = 1

    # set zero marker
    ret.append([last_time, 0])

    # for all range boundaries:
    #     if time changed:
    #         store prervious tuple
    #     if a range STARTed at this time, increase concurrency,
    #     if a range ENDed   at this time, decrease concurrency,
    # store trailing tuple

    for time_, mode in times[1:]:
        if time_ != last_time:
            ret.append([last_time, concurrency])
            last_time = time_

        if mode == START: concurrency += 1
        else            : concurrency -= 1

    ret.append([last_time, concurrency])

    assert concurrency == 0, \
            'inconsistent range structure? %s : %s : %s' % (ranges, ret, times)

    return ret


# ------------------------------------------------------------------------------
#
def partition(space, nparts):
    '''
    create balanced partitions from an iterable space.  This method preserves
    contiguous order.

    kudos:
    http://code.activestate.com/recipes/425397-split-a-list-into-roughly-equal-sized-pieces/
    '''

    n   = len(space)
    b   = 0
    ret = list()

    for k in range(nparts):

        q, r = divmod(n - k, nparts)
        a, b = b, b + q + (r != 0)

        ret.append(space[a:b])

    return ret


# ------------------------------------------------------------------------------
#
def in_range(value, ranges):
    '''
    checks if a float value `value` is in any of the given `ranges`, which are
    assumed to be a list of tuples (or a single tuple) of start end end points
    (floats).
    Returns `True` or `False`.
    '''

    # is there anythin to check?
    if not ranges or not len(ranges):
        return False

    if not isinstance(ranges[0], list):
        ranges = [ranges]

    for r in ranges:
        if value >= r[0] and value <= r[1]:
            return True

    return False


# ------------------------------------------------------------------------------
#
def remove_common_prefix(data, extend=0):
    '''
    For the given list of strings, remove the part which appears to be a common
    prefix to all of them.  If `extend` is given, the last `n` letters of the
    prefix are preserved.

    See also https://stackoverflow.com/questions/46128615/ .
    '''

    if not data:
        return data

    # sort the list, get prefix from comparing first and last element
    s1 = min(data)
    s2 = max(data)
    prefix = s1

    for i, c in enumerate(s1):
        if c != s2[i]:
            prefix = s1[:i]
            break

    if extend > 0 and len(prefix) < len(s1):
        prefix = s1[:min(len(prefix) - extend, len(s1))]

    # remove the found prefix from all elements
    return [elem.split(prefix, 1)[-1] for elem in data]


# ------------------------------------------------------------------------------
#
# Assume we want to schedule a set of tasks over a set of resources.  To
# maximize utilization, we want to place the largest tasks first and fill the
# gap with small tasks, thus we sort the list of tasks by size and schedule
# beginning with the largest.
#
# For large number of tasks this can become expensive, specifically once the set
# of resources is near exhaustion: we need to search through the whole task list
# before being able to, possibly, place some small tasks.
#
# The algorithm below attempts to use the knowledge that, once a task of
# a certain size can *not* be scheduled, no larger task can.  It performs as
# follows:
#
#       1  attempt to schedule the largest task (n)
#       2  if success
#       3      schedule the next smaller task (n = n - 1)
#       4      if success
#       5          goto (2)
#       7  bisect distance to smallest task: l = bisect(m, 0)
#       8  schedule task l
#       9  if success   # task l is small enough
#      10      bisect distance toward larger task: l = bisect(m, l)
#      11      goto 9
#      12  else         # task l is still too large
#      13      bisect distance towards smaller tasks: l = bisect(l, 0)
#      14      goto 9
#
# This algorithm stops once (m - l == 1), which is the boundary between 'too
# large' and 'small enough' tasks.  From here, we mark all tasks > m as too
# large, set n = l, and begin from scratch.
#
# The implementation below contains some further optimizations.  Specifically,
# we remember the schedule results of all bisections, so that we don't schedule
# or test those tasks twice.  Further, we assume that scheduling a task may
# alter the result for all further schedule attempts (apart from tasks we know
# are too large).
#
# Example:
#
#   # --------------------------------------------------------------------------
#   #
#   # Tasks larger 65 cannot be scheduled.  Once 65 is scheduled, 64 is too
#   # large and fails.  63, 62, 61 fit, but then space is tight and all tasks up
#   # to 42 fail.  42 to 22 get scheduled ok, but then space is out again for
#   # everything larger than 7 - the last 8 tasks (very small) can be scheduled.
#   #
#   # Out of the 128 tasks, 32 will get schedules, so 96 will not.  Instead of
#   # trying schedule those 96 tasks, the algorithm will only test 15 tasks
#   # unsuccessfully (~15%).  The effect improves with larger numbers of tasks
#   # and less available resources, which is usually the case when the RP
#   # scheduler sees high contention between scheduling and unscheduling
#   # operations.  For example, the same setup for 128k tasks results in 25
#   # failed scheduling attempts (0.01%).
#   #
#   failed = list()
#   def schedule(n):
#       if n in [61, 62, 63, 65]    or \
#          n in list(range(22, 42)) or \
#          n < 8:
#           scheduled.append(n)
#           return True   # could be scheduled
#       else:
#           failed.append(n)
#           return False  # schedule failed
#
#   tasks = list(range(128 * 1024))
#   good, bad = lazy_bisect(tasks, schedule)
#
#   assert len(failed) == 25)
#   for task in good:
#       assert schedule(task) is True
#       task not in bad
#   for task in bad:
#       assert schedule(task) is False
#       task not in good
#   # --------------------------------------------------------------------------
#
def lazy_bisect(data, check,
                on_ok=None, on_nok=None, on_fail=None, on_skip=None,
                ratio=0.5, log=None):
    '''
    Find the next potentially good candidate element in a presumably ordered
    list `data` (from smallest to largest). The given callable `check` should
    accept a list element for good/bad checks, and return `True` for good (small
    enough), `False` otherwise (too large).

    Note that this is only a bisect if `ratio` is set at `0.5` - otherwise the
    search will back off through the list faster (1.0 >= ratio > 0.5) or slower
    (0.5 > ratio >= 0.0) than bisect.

    The method returns three lists, one list with elements of data which were
    checked `good`, and one list of elements checked as `bad`.  Checks which
    raise an exception are also handled as `bad`, any element on which this
    happens will be returned in the third list (`fail`).  Note that this last
    list (`fail`) will consist of tuples: [`element`, `error`] where `error` is
    the exception's error message.

    `on_ok` will be called on tasks where `check()` returned `good`, `on_nok`
    where `check()` returned `bad`, `on_fail()` where `check()` raised an
    exceptionm and `on_skip` where bisect decided that no
    `check()` invocation is needed.  The respective data item is the only
    argument to the calls.

    '''

    if not data:
        return [], [], []

    if not log:
        log = Logger('radical.utils.alg')

    if ratio > 1.0: ratio = 1.0
    if ratio < 0.0: ratio = 0.0

    last_good  = None    # last known good index
    last_bad   = None    # last known bad index
    check_good = list()  # found good index
    check_bad  = list()  # found bad index
    check_fail = list()  # found bad index

    # --------------------------------------------------------------------------
    def wrapcheck(thing, skip=False):
        ret = None
        if skip:
            if on_skip:
                on_skip(thing)
        else:
            ret = check(thing)
            if ret is True  and on_ok : on_ok(thing)
            if ret is False and on_nok: on_nok(thing)

        return ret
    # --------------------------------------------------------------------------

  # # --------------------------------------------------------------------------
  # def state_hay():
  #     hay = ''
  #     for i,x in enumerate(data):
  #         if not i % 10       : hay += '|'
  #         if check(x) is True : hay += '#'
  #         else                : hay += '.'
  #     if not hay.endswith('|'):
  #         hay += '|'
  #     log.debug('%30s %s', '', hay)
  # # --------------------------------------------------------------------------
  #
  # # --------------------------------------------------------------------------
  # def state_needle(msg=''):
  #     needle = ''
  #     for i,x in enumerate(data):
  #         if not i % 10       : needle += '|'
  #         if   x in check_good: needle += '#'
  #         elif x in check_bad : needle += '.'
  #         else                : needle += ' '
  #     if not needle.endswith('|'):
  #         needle += '|'
  #     g = last_good
  #     b = last_bad
  #     if g is None: g = '?'
  #     if b is None: b = '?'
  #     log.debug('%3s - %3s %20s %s %s', g, b, ' ', needle, msg)
  # # --------------------------------------------------------------------------

  # state_hay()
    first = True
    while True:

        # if we don't know anything, yet, check the first element
        if  last_good is None and \
            last_bad  is None:

            assert first
            first = False

            idx = len(data) - 1

            try:
                ret = wrapcheck(data[idx])

                if ret is True: check_good.append(idx)
                else          : check_bad .append(idx)
            except Exception as e:
                ret = False
                check_fail.append([idx, str(e)])

          # state_needle('start %s %s' % (idx, ret))

            if ret is True: last_good = idx
            else          : last_bad  = idx


        # If we only know a good one, just try the next in the list.  This can
        # happen repeatedly, and we'll traverse the list for long as we can
        elif last_good is not None and \
             last_bad  is     None:

            # make sure we still have something to check
            if last_good == 0:
                break

            # check the next element
            idx = last_good - 1

            if idx in check_good:
                last_good = idx
              # state_needle('known %3d True' % idx)
                continue

            if idx in check_bad:
                last_bad = idx
              # state_needle('known %3d False' % idx)
                continue

            try:
                ret = wrapcheck(data[idx])

                if ret is True: check_good.append(idx)
                else          : check_bad .append(idx)
            except Exception as e:
                ret = False
                check_fail.append([idx, str(e)])

            if ret is True: last_good = idx
            else          : last_bad  = idx  # break out of this branch

          # state_needle('good  %3d %s' % (idx, ret))


        # If we know a bad one, we bisect the remaining list and check the
        # found candidate.  If it is also bad, we assume all intermediate
        # elements as bad.  If it is good, we set `last_good` and try again.
        #
        # If last_good is not known here, we bisect to begin of list (0),
        # otherwise we bisect to last_good.
        elif last_bad is not None:

            # make sure we still have something to check
            if last_bad == 0:
                break

            if last_good is not None:
                if last_good > last_bad:
                    last_good = None

            # bisect for next candidate
            if last_good is not None:

                # bisect the difference, using the given ratio
                idx = int(m.ceil((last_bad - last_good + 1) * ratio + last_good))
                idx = min(idx, last_bad - 1)

            else:
                # bisect to begin of data
                idx = int(m.ceil((last_bad + 1) * ratio))
                idx = min(idx, last_bad - 1)

          # state_needle('range %3d?' % idx)

            # make sure we make progress: if space is too small for bisect, then
            # increase last_good or decrease last_bad
            if idx == last_good: idx = last_good + 1
            if idx == last_bad : idx = last_bad  - 1

            if idx < 0: idx = 0

            # check this bisected index (if we don't know it yet)
            if   idx in check_good: ret = True
            elif idx in check_bad:  ret = False
            else:
                try:
                    ret = wrapcheck(data[idx])

                    if ret is True: check_good.append(idx)
                    else          : check_bad .append(idx)
                except Exception as e:
                    ret = False
                    check_fail.append([idx, str(e)])

          # state_needle('range %3d %s' % (idx, ret))

            if ret is True:
                # found a new good one closer to the last bad one - bisect again
                # until we neighbor last_bad.  In that case, reset the last_bad
                # and restart searching (and bisecting) from that meeting point.
                last_good = idx

                if last_bad is not None:
                    if last_bad < last_good:
                        last_good = None
                      # state_needle('range  A')

                    elif last_bad - last_good == 1:
                        last_bad = None
                      # state_needle('range  a')

            else:

                # found a new, smaller bad one - consider all indexes between
                # id and the previous last_bad as bad
                for i in range((last_bad - idx - 1)):
                    this = last_bad - i - 1
                    if this not in check_bad and \
                       this not in check_good:
                        wrapcheck(data[this], skip=True)
                        check_bad.append(this)
              # state_needle('Range')
                last_bad = idx

                if last_good is not None and last_good > last_bad:
                    # this last_good is not interesting anymore  we search
                    # downwards
                    last_good = None

  # state_hay()

    assert len(data) == len(check_good) + len(check_bad) + len(check_fail)

    return [ data[i]     for i   in check_good], \
           [ data[i]     for i   in check_bad],  \
           [[data[i], e] for i,e in check_fail]


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test = [[ 0, 10],
            [20, 30],
            [40, 50],
            [60, 70],
            [80, 90],
            [ 5, 15],
            [35, 55]]

    import pprint
    pprint.pprint(test)
    pprint.pprint(collapse_ranges(test))

    test_space = list(range(75))
    parts = partition(test_space, 8)
    for part in parts:
        print('%3d: %s' % (len(part), part))


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    import time
    scheduled = list()
    def schedule(n, check=False):

        if not check:
            time.sleep(0.1)  # scheduling is slow (but not checks)
            print(n)

        if n in [61, 62, 63, 65]    or \
           n in list(range(22, 42)) or \
           n < 8:
            scheduled.append(n)
            if not check: print('ok')
            return True

        else:
            if not check: print('--')
            return False

    tasks = list(range(128 * 1024))
    tasks = lazy_bisect(tasks, schedule)

    for task in tasks:
        assert schedule(task, check=True) is False

    for task in scheduled:
        assert task not in tasks


# ------------------------------------------------------------------------------





__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2021-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

# ------------------------------------------------------------------------------
#
# We provide a base class for all kinds of dict-like data objects in the RCT
# stack: resource configs, agent configs, job descriptions, task descriptions,
# pilot descriptions, workload descriptions etc.
#
# Class provides:
#
#   - dict like API
#   - public dict keys exposed as attributes ("Munch" approach)
#   - schema based type definitions
#   - optional runtime type checking
#

import copy
import sys

from .misc      import as_list, as_tuple, is_string


# ------------------------------------------------------------------------------
#
class TDErrorMixin:

    # --------------------------------------------------------------------------
    #
    def __init__(self, msg=None, level=1):

        f = sys._getframe(level)
        if 'self' in f.f_locals:
            cls_name = type(f.f_locals['self']).__name__
        elif 'cls' in f.f_locals:
            cls_name = f.f_locals['cls'].__name__
        else:
            cls_name = '<>'

        super().__init__('%s.%s%s' % (cls_name,
                                      f.f_code.co_name,
                                      msg and ' - %s' % msg or ''))


# ------------------------------------------------------------------------------


class TDError(TDErrorMixin, Exception):
    pass


class TDKeyError(TDErrorMixin, KeyError):
    pass


class TDTypeError(TDErrorMixin, TypeError):
    pass


class TDValueError(TDErrorMixin, ValueError):
    pass


# ------------------------------------------------------------------------------
#
class TypedDictMeta(type):

    # --------------------------------------------------------------------------
    #
    def __new__(mcs, name, bases, namespace):

        # guaranteed class "base" attributes
        _base_namespace = {
            '_schema'      : {},
            '_defaults'    : {},
            '_self_default': False,  # convert unschemed-dict into class itself
            '_check'       : False,  # attribute type checking on set
            '_cast'        : True,   # attempt to cast on type mismatch
            '_deep'        : True    # use deepcopy on defaults
        }

        for _cls in bases:
            deep = namespace.get('_deep', _base_namespace['_deep'])
            for k in _base_namespace.keys():
                _cls_v = getattr(_cls, k, None)
                if _cls_v is not None:
                    if   k == '_schema':
                        _base_namespace[k].update(_cls_v)
                    elif k == '_defaults':
                        if deep:
                            _base_namespace[k].update(copy.deepcopy(_cls_v))
                        else:
                            _base_namespace[k].update(_cls_v)
                    else:
                        _base_namespace[k] = _cls_v

        for k, v in _base_namespace.items():
            if isinstance(v, dict):
                v.update(namespace.get(k, {}))
                namespace[k] = v
            elif k not in namespace:
                namespace[k] = v

        _new_cls = super().__new__(mcs, name, bases, namespace)

        if _new_cls.__base__ is not dict:

            # register sub-classes
            from .serialize import register_serializable
            register_serializable(_new_cls)

        return _new_cls



# ------------------------------------------------------------------------------
#
class TypedDict(dict, metaclass=TypedDictMeta):

    # --------------------------------------------------------------------------
    #
    def __init__(self, from_dict=None, **kwargs):
        '''
        Create a typed dictionary (tree) from `from_dict`.

        from_dict: data to be used for initialization

        NOTE: the names listed below are valid keys when used via the
              dictionary API, and can also be *set* via the property API, but
              they cannot be *queried* via the property API as their names
              conflict with the class method names:

                  as_dict
                  clear
                  get
                  items
                  keys
                  pop
                  popitem
                  setdefault
                  update
                  values
                  verify

              Names with a leading underscore are not supported.

              Supplied `from_dict` and kwargs are used to initialize the object
              data -- the `kwargs` take preceedence over the `from_dict` if both
              are specified (note that `from_dict` and `self` are invalid
              `kwargs`).
        '''

        from .serialize import register_serializable

        register_serializable(self.__class__)


        if self._deep:
            self.update(copy.deepcopy(self._defaults))
        else:
            self.update(self._defaults)

        self.update(from_dict)

        if kwargs:
            self.update(kwargs)


    # --------------------------------------------------------------------------
    #
    def update(self, other):
        '''
        Overload `dict.update()`: the call is used to ensure that sub-dicts are
        instantiated as their respective TypedDict-inheriting class types,
        if so specified by the respective schema.

        So, if the schema contains::

          {
            ...
            'foo' : BarTypedDict
            'foos': [BarTypedDict]
            ...
          }

        where `BarTypedDict` is a valid type in the scope of the schema
        definition, which inherits from `ru.TypedDict`, then `update()` will
        ensure that the value for key `foo` is indeed of type `ru.TypedDict`.
        An error will be raised if (a) `BarTypedDict` does not have a single
        parameter constructor like `ru.TypedDict`, or (b) the `data` value for
        `foo` cannot be used as `from_dict` parameter to the `BarTypedDict`
        constructor.
        '''
        if not other:
            return

        for k, v in other.items():
            if isinstance(v, dict):
                t = self._schema.get(k) or \
                    (type(self) if self._self_default else TypedDict)
                if isinstance(t, type) and issubclass(t, TypedDict):
                    # cast to expected TypedDict type
                    if not self.get(k):
                        self[k] = t()
                    self[k].update(v)
                    continue
            self[k] = v


    # --------------------------------------------------------------------------
    #
    def __deepcopy__(self, memo):
        '''
        return a new instance of the same type, not an original TypedDict,
        otherwise if an instance of TypedDict-based has an attribute of other
        TypedDict-based type then `verify` method will raise an exception
        '''
        return type(self)(from_dict=copy.deepcopy(self._data))


    # --------------------------------------------------------------------------
    #
    # base functionality to manage items
    #
    def __getitem__(self, k):
        return self._data[k]

    def __setitem__(self, k, v):
        self._data[k] = self._verify_setter(k, v)

    def __delitem__(self, k):
        del self._data[k]

    def __contains__(self, k):
        return k in self._data

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def clear(self):
        self._data.clear()


    # --------------------------------------------------------------------------
    #
    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
            return default
        return self[key]

    def pop(self, key, default=None):
        if key in self:
            value = self[key]
            del self[key]
            return value
        elif default is not None:
            return default
        else:
            raise TDKeyError('key "%s" not found' % key)

    def popitem(self):
        if len(self):
            key = list(self.keys())[-1]  # LIFO if applicable
            value = self[key]
            del self[key]
            return key, value
        else:
            raise TDError('no data')


    # --------------------------------------------------------------------------
    #
    # base functionality for attribute access
    #
    def __getattr__(self, k):

        if k == '_data':
            if '_data' not in self.__dict__:
                self.__dict__['_data'] = dict()
            return self.__dict__['_data']

        if k.startswith('__'):
            return object.__getattribute__(self, k)

        data   = self._data
        schema = self._schema

        if   not  schema: return data.get(k)
        elif k in schema: return data.get(k)
        else            : return data[k]

    def __setattr__(self, k, v):

        if k.startswith('__'):
            return object.__setattr__(self, k, v)

        self._data[k] = self._verify_setter(k, v)

    def __delattr__(self, k):

        if k.startswith('__'):
            return object.__delattr__(self, k)

        del self._data[k]


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return '%s: %s' % (type(self).__qualname__, str(self))


    # --------------------------------------------------------------------------
    #
    def as_dict(self, _annotate=False):

        tgt = dict()

        for k, v in self._data.items():

            if isinstance(v, TypedDict):
                tgt[k] = v.as_dict(_annotate=_annotate)
            else:
                tgt[k] = as_dict(v, _annotate=_annotate)
            if _annotate:
                tgt['_type'] = type(self).__name__

        return tgt


    # --------------------------------------------------------------------------
    #
    @classmethod
    def __raise_type_error(cls, k, v, t, level=4):
        raise TDTypeError(
            'attribute "%s" - expected type %s, got %s' % (k, t, type(v)),
            level=level)

    @classmethod
    def _verify_base(cls, k, v, t):
        if cls._cast:
            try:
                return t(v)
            except (TypeError, ValueError):
                pass
        cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_bool(cls, k, v, t):
        if cls._cast:
            if str(v).lower() in ['true', 'yes', '1']:
                return True
            if str(v).lower() in ['false', 'no', '0']:
                return False
        cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_tuple(cls, k, v, t):
        if cls._cast:
            v = as_tuple(v)
            return tuple(
                [cls._verify_kvt(k + ' tuple element', _v, t[0]) for _v in v])
        else:
            if isinstance(v, tuple):
                return v
            cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_list(cls, k, v, t):
        if cls._cast:
            v = as_list(v)
            return [cls._verify_kvt(k + ' list element', _v, t[0]) for _v in v]
        else:
            if isinstance(v, list):
                return v
            cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_dict(cls, k, v, t):
        if cls._cast:
            t_k = list(t.keys())[0]
            t_v = list(t.values())[0]
            return {cls._verify_kvt(_k, _k, t_k):
                    cls._verify_kvt(_k, _v, t_v)
                    for _k, _v in v.items()}
        else:
            if issubclass(type(v), dict):
                return v
            cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_typeddict(cls, k, v, t):
        if cls._cast:
            if issubclass(type(v), t): return v.verify()
            # different TypedDict-base, but has a subset of schema
            if isinstance(v, dict)   : return t(from_dict=v).verify()
        else:
            if issubclass(type(v), t):
                return v
        cls.__raise_type_error(k, v, t)

    @classmethod
    def _verify_kvt(cls, k, v, t):
        if t is None or v is None      : return v
        if isinstance(t, type):
            if issubclass(t, TypedDict): return cls._verify_typeddict(k, v, t)
            # check base types
            if isinstance(v, t)        : return v
            if t in [str, int, float]  : return cls._verify_base(k, v, t)
            if t is bool               : return cls._verify_bool(k, v, t)
            cls.__raise_type_error(k, v, t, level=3)
        if isinstance(t, tuple)        : return cls._verify_tuple(k, v, t)
        if isinstance(t, list)         : return cls._verify_list(k, v, t)
        if isinstance(t, dict)         : return cls._verify_dict(k, v, t)
        if cls._cast                   : return v
        raise TDTypeError('no verifier defined for type %s' % t, level=2)

    def verify(self):

        if self._schema:
            for k, v in self._data.items():

                if k.startswith('__'):
                    continue

                if k not in self._schema:
                    raise TDKeyError('key "%s" not in schema' % k)

                self._data[k] = self._verify_kvt(k, v, self._schema[k])
        self._verify()
        return self


    # --------------------------------------------------------------------------
    #
    def _verify_setter(self, k, v):

        if   not self._check : return v
        elif not self._schema: return v

        if k not in self._schema:
            raise TDKeyError('key "%s" not in schema' % k, level=2)
        return self._verify_kvt(k, v, self._schema[k])


    # --------------------------------------------------------------------------
    #
    def _verify(self):
        '''
        Can be overloaded
        '''
        pass


    # --------------------------------------------------------------------------
    #
    def _query(self, key, default=None, last_key=True):
        '''
        For a query like

            typeddict.query('some.path.to.key', 'foo')

        this method behaves like:

            typeddict['some']['path']['to'].get('key', default='foo')

        flag `last_key` allows getting `default` value if any sub-key is missing
        (by default only if the last key is missing then return `default` value)
        '''
        if not key:
            raise TDKeyError('empty key on query')
        key_seq = key.split('.') if is_string(key) else list(key)

        output = self
        while key_seq:

            sub_key = key_seq.pop(0)

            if sub_key in output:
                output = output[sub_key]
                # if there are more sub-keys in a key sequence
                if key_seq and not isinstance(output, dict):
                    raise TDValueError(
                        'value for sub-key "%s" is not of dict type' % sub_key)

            elif not key_seq or not last_key:
                output = default
                break

            else:
                raise TDKeyError('intermediate key "%s" missed' % sub_key)

        return output


# ------------------------------------------------------------------------------
#
def as_dict(src, _annotate=False):
    '''
    Iterate given object, apply `as_dict()` to all typed
    values, and return the result (effectively a shallow copy).
    '''
    if isinstance(src, TypedDict):
        return src.as_dict(_annotate=_annotate)

    if isinstance(src, dict):
        return {k: as_dict(v, _annotate) for k, v in src.items()}

    if isinstance(src, list):
        return [as_dict(x, _annotate) for x in src]

    if isinstance(src, tuple):
        return tuple([as_dict(x, _annotate) for x in src])

    return src


# ------------------------------------------------------------------------------





import os
import pprint

import radical.utils as ru

# ------------------------------------------------------------------------------
# a default config, which defines:
#   - the structure / keys of the config settings
#   - the default values for the keys
app_config = {
    'log_level'    : 0,
    'scheduler'    : 'RoundRobin',
    'resources'    : ['tutorial.radical.org', 'localhost'],
    'resource_cfg' :
    {
        'localhost' :
        {
            'needs_account' : False,
            'supports_mpi'  : False,
            'username'      : None,
            'account'       : None
        },
        'tutorial.radical.org' :
        {
            'needs_account' : False,
            'supports_mpi'  : True,
            'username'      : 'tut_007',
            'account'       : None
        },
        'sierra.futuregrid.org' :
        {
            'needs_account' : False,
            'supports_mpi'  : True,
            'username'      : None,
            'account'       : None
        },
        'india.futuregrid.org' :
        {
            'needs_account' : False,
            'supports_mpi'  : True,
            'username'      : None,
            'account'       : None
        },
        'stampede.tacc.utexas.edu' :
        {
            'needs_account' : True,
            'supports_mpi'  : True,
            'username'      : None,
            'account'       : None
        }
    }
}

# ------------------------------------------------------------------------------
# location of the user config
# the config could contain:
#
# {
#     "scheduler"    : "Backfilling",
#     "resources"    : ["india.furturegrid.org", "sierra.futuregrid.org"],
#     "resource_cfg" :
#     {
#         "*.futuregrid.org" :
#         {
#             "username"      : "merzky"
#         }
#     }
# }
USER_CONFIG_PATH = os.environ.get ('HOME', '/tmp') + '/.my_app.cfg'

# load the user config, and merge it with the default config
user_config = ru.read_json_str (USER_CONFIG_PATH)


# merge the user config into the app config, so that the user config keys are
# applied where appropriate
ru.dict_merge (app_config, user_config, policy='overwrite', wildcards=True)


# lets see what we got
pprint.pprint (app_config)


# this should result in :
#
# {
#     'log_level'   : 0,
#     'scheduler'   : 'Backfilling',
#     'resources'   : ['india.furturegrid.org', 'sierra.futuregrid.org'],
#     'resource_cfg':
#     {
#         '*.futuregrid.org':
#         {
#             'username'     : 'merzky'
#         },
#         'india.futuregrid.org':
#         {
#             'account'      : None,
#             'needs_account': False,
#             'supports_mpi' : True,
#             'username'     : 'merzky'
#         },
#         'localhost':
#         {
#             'account'      : None,
#             'needs_account': False,
#             'supports_mpi' : False,
#             'username'     : None
#         },
#         'sierra.futuregrid.org':
#         {
#             'account'      : None,
#             'needs_account': False,
#             'supports_mpi' : True,
#             'username'     : 'merzky'
#         },
#         'stampede.tacc.utexas.edu':
#         {
#             'account'      : None,
#             'needs_account': True,
#             'supports_mpi' : True,
#             'username'     : None
#         },
#         'tutorial.radical.org':
#         {
#             'account'      : None,
#             'needs_account': False,
#             'supports_mpi' : True,
#             'username'     : 'tut_007'
#         }
#     }
# }






import time
import queue

import threading as mt

from collections import defaultdict
from typing      import List, Union

from ..misc       import as_list
from ..ids        import generate_id
from ..logger     import Logger

from .flux_module import FluxModule


# ------------------------------------------------------------------------------
#
class FluxHelperV1(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, uri : str,
                       log : Logger = None) -> None:

      # print('=== v1 flux helper ===')

        self._uri      = uri
        self._log      = log or Logger('radical.utils.flux')
        self._uid      = generate_id('ru.flux')

        self._fm       = FluxModule()
        self._handle   = self._fm.core.Flux(self._uri)
        self._api_lock = mt.Lock()

        # journal watcher
        self._jthread  = None
        self._jterm    = mt.Event()

        # event handle thread
        self._ethread  = None
        self._equeue   = queue.Queue()

        # submit thread
        self._sthread  = None
        self._squeue   = queue.Queue()
        self._sevent   = mt.Event()

        self._idlock   = mt.Lock()          # lock ID dicts
        self._elock    = mt.Lock()          # lock event dict
        self._task_ids = dict()             # flux ID -> task ID
        self._flux_ids = dict()             # task ID -> flux ID
        self._events   = defaultdict(list)  # flux ID -> event list
        self._cbacks   = list()             # list of callbacks

        self._fm.verify()
        for line in (self._fm.version or '').splitlines():
            self._log.info('flux version: %s', line)


    # --------------------------------------------------------------------------
    #
    def start(self, launcher: str   = None) -> None:

        with self._api_lock:

            if self._jthread is not None:
                return

            self._jthread = mt.Thread(target=self._jwatcher)
            self._jthread.daemon = True
            self._jthread.start()

            self._ethread = mt.Thread(target=self._ewatcher)
            self._ethread.daemon = True
            self._ethread.start()

            self._sthread = mt.Thread(target=self._swatcher)
            self._sthread.daemon = True
            self._sthread.start()


    # --------------------------------------------------------------------------
    #
    def stop(self):

        with self._api_lock:

            if self._handle is None:
                self._jterm.set()
                self._jthread.join()

                # FIXME: shutdown flux instance
                self._flux_service = None
                self._uri          = None
                self._handle       = None


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self) -> str:
        return self._uid

    @property
    def uri(self) -> str:
        return self._uri


    # --------------------------------------------------------------------------
    #
    def _jwatcher(self):

        # NOTE: *never* used self._handle in this thread, as it is not thread
        #      safe.  Instead, use the private handle created here
        fh = self._fm.core.Flux(self._uri)

        # start watching the event journal
        journal = self._fm.job.JournalConsumer(fh)
        journal.start()

        while not self._jterm.is_set():

            try:
                event = journal.poll(timeout=1.0)
                if event:
                    # FIXME: How can that ever *not* be a journal event?
                    #        But it has happened...
                    self._handle_events(fh, event.jobid, event)

            except TimeoutError:
                pass


    # --------------------------------------------------------------------------
    #
    def _swatcher(self):

        self._log.debug('swatcher started')

        # if we get new specs, submit them, return IDs to iqueue, and also
        # forward ID to ewatcher
        fh = self._fm.core.Flux(self._uri)
        while True:

            try:
                specs = self._squeue.get(block=True, timeout=1.0)
                self._log.debug('got %d specs', len(specs))

            except queue.Empty:
                continue

            except:
                self._log.exception("exception")
                raise

            with self._idlock:

                try:

                    futs = list()
                    for spec in specs:
                        tid = spec.attributes['user']['uid']
                        fut = self._fm.job.submit_async(fh, spec, waitable=True)
                        futs.append([fut, tid])

                    for fut, tid in futs:
                        fid = fut.get_id()
                        self._task_ids[fid] = tid
                        self._flux_ids[tid] = fid

                        # trigger an event check
                        self._equeue.put(fid)

                except Exception:
                    self._log.exception("exception")
                    raise

                finally:
                    # trigger submit completion
                    self._log.debug('submit done')
                    self._sevent.set()


    # --------------------------------------------------------------------------
    #
    def _ewatcher(self):

        # if we get a new job ID, check if we have events for it

        fh = self._fm.core.Flux(self._uri)
        while True:

            try:
                fid = self._equeue.get(timeout=1.0)
                self._handle_events(fh, fid)

            except queue.Empty:
                continue


    # --------------------------------------------------------------------------
    #
    def register_cb(self, cb: callable) -> None:

        with self._api_lock, self._elock:
            self._log.debug('register cb %s', cb)
            self._cbacks.append(cb)


    # --------------------------------------------------------------------------
    #
    def unregister_cb(self, cb: callable) -> None:

        with self._api_lock, self._elock:
            self._cbacks.remove(cb)


    # --------------------------------------------------------------------------
    #
    def _handle_events(self, fh   : 'flux.Flux',
                             fid  : 'flux.job.JobID',
                             event: 'flux.job.journal.JournalEvent' = None
                      ) -> None:

        with self._elock:

          # self._log.debug_9('event %s: %s', fid, event)

            # if triggered by submit, check if we have anything to do
            if not event:
                if fid not in self._events:
                    return

            # check if we can handle the event - otherwise store it
            if not self._cbacks:
                self._events[fid].append(event)
                return

            # check if application knows the task - otherwise store the event
            if fid not in self._task_ids:
                self._events[fid].append(event)
                return

            tid = self._task_ids[fid]

            # task is known, flush stored events
            for ev in self._events[fid]:
                for cb in self._cbacks:
                    try   : cb(tid, ev)
                    except: self._log.exception('cb failed')
                self._events[fid] = []

            # process the current event
            if event:
                for cb in self._cbacks:
                    try   : cb(tid, event)
                    except: self._log.exception('cb failed')


    # --------------------------------------------------------------------------
    #
    def submit(self, specs: List['flux.job.JobspecV1']) -> List[str]:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            self._log.debug('== submit %d specs start', len(specs))
            tids = [spec.attributes['user']['uid'] for spec in specs]

            self._sevent.clear()
            self._squeue.put(specs)
            self._sevent.wait(timeout=60.0)

            if not self._sevent.is_set():
                raise RuntimeError('flux submit timeout')

            self._log.debug('== submit %d specs done', len(specs))

            return tids


    # --------------------------------------------------------------------------
    #
    def cancel(self, tids: [Union[str, List[str]]]) -> None:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            with self._idlock:
                for tid in as_list(tids):
                    fid = self._flux_ids[tid]
                    self._fm.job.cancel_async(self._handle, fid, reason='user cancel')


    # --------------------------------------------------------------------------
    #
    def wait(self, tids: [Union[str, List[str]]]) -> None:

        with self._api_lock:

            if not self._handle:
                raise RuntimeError('flux instance not started')

            tids = as_list(tids)
            with self._idlock:
                fids = [self._flux_ids[tid] for tid in tids]

            for fid in fids:
                self._fm.job.wait(self._handle, fid)

            # FIXME: remove tasks which have been waited for.


# ------------------------------------------------------------------------------





import time

import threading as mt

from rc.process  import Process
from functools   import partial
from typing      import List

from ..url        import Url
from ..ids        import generate_id
from ..logger     import Logger

from .flux_module import FluxModule


# ------------------------------------------------------------------------------
#
class FluxService(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, uid     : str    = None,
                       log     : Logger = None,
                       launcher: str    = None
                       ) -> None:

        self._uid      = uid      or generate_id('ru.flux')
        self._log      = log      or Logger('radical.utils.flux', level='DEBUG')
        self._launcher = launcher or ''

        self._fm    = FluxModule()
        self._uri   = None
        self._r_uri = None
        self._host  = None
        self._proc  = None
        self._ready = mt.Event()

        self._fm.verify()


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self) -> str:
        return self._uid

    @property
    def uri(self) -> str:
        return self._uri


    @property
    def r_uri(self) -> str:
        return self._r_uri


    # --------------------------------------------------------------------------
    #
    def _proc_line_cb(self, prefix: str,
                            proc  : Process,
                            lines : List[str]
                     ) -> None:

        try:
            for line in lines:
                self._log.info('%s: flux io : %s', self._uid, line)

                if line.startswith('FLUX_URI='):
                    parts = line.strip().split(' ', 1)
                    self._log.info('%s: found flux info: %s', self._uid, parts)

                    self._uri   = parts[0].split('=', 1)[1]
                    self._host  = parts[1].split('=', 1)[1]

                    url         = Url(self._uri)
                    url.host    = self._host
                    url.schema  = 'ssh'
                    self._r_uri = str(url)

                    self._log.info('%s: flux uri: %s', self._uid, self._uri)
                    self._log.info('%s:    r uri: %s', self._uid, self._r_uri)
                    self._ready.set()
        except:
            self._log.exception('line processing failed')


    # --------------------------------------------------------------------------
    #
    def _proc_state_cb(self, proc: Process, state: str) -> None:

        self._log.info('flux instance state update: %s', state)
        if state in Process.FINAL:

            self._log.info('flux instance stopped: %s', state)
            self.stop()


    # --------------------------------------------------------------------------
    #
    def start(self, timeout: float = None) -> None:

        fcmd  = 'echo FLUX_URI=\\$FLUX_URI FLUX_HOST=\\$(hostname) '
        fcmd += ' && flux resource list '
        fcmd += ' && sleep inf '
        cmd   = ' %s start bash -c "%s"' % (self._fm.exe, fcmd)

        if self._launcher:
            cmd = '%s %s' % (self._launcher, cmd)

        self._log.info('%s: start flux instance: %s', self._uid, cmd)

        p = Process(cmd)
        p.register_cb(p.CB_OUT_LINE, partial(self._proc_line_cb, 'out'))
        p.register_cb(p.CB_ERR_LINE, partial(self._proc_line_cb, 'err'))
        p.register_cb(p.CB_STATE, self._proc_state_cb)
        p.polldelay = 0.1
        p.start()

        self._proc  = p
        self._ptime = time.time()

        return self.ready(timeout=timeout)


    # --------------------------------------------------------------------------
    #
    def ready(self, timeout: float = None) -> None:

        if timeout is not None:
            if timeout < 0: self._ready.wait()
            else          : self._ready.wait(timeout)

        return self._ready.is_set()


    # --------------------------------------------------------------------------
    #
    def stop(self) -> None:

        if not self._proc:
            self._uri  = None
            return

        self._proc.cancel()
        self._proc.wait()

        self._uri  = None
        self._proc = None

        self._log.info('%s: found flux uri: %s', self._uid, self.uri)


# ------------------------------------------------------------------------------




# pylint: disable=unused-import,import-error,no-name-in-module,eval-used,unspecified-encoding

__author__    = 'Radical.Utils Development Team (Andre Merzky)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import re
import os
import tempfile

from .serialize import to_json, from_json
from .misc      import as_string, ru_open


# ------------------------------------------------------------------------------
#
def read_json(fname, filter_comments=True):
    '''
    Comments line in the form of

        # some json data or text

    are stripped from json before parsing:

        import pprint
        pprint.pprint(read_json("my_file.json"))
    '''

    with ru_open(fname) as f:

        try:
            output = parse_json(f.read(), filter_comments)
        except ValueError as e:
            raise ValueError('error parsing %s: %s' % (fname, e)) from e

    return output


# ------------------------------------------------------------------------------
#
def read_json_str(filename, filter_comments=True):
    '''
    same as read_json, but converts unicode strings and byte arrays to ASCII
    strings.
    '''

    return as_string(read_json(filename, filter_comments))


# ------------------------------------------------------------------------------
#
def write_json(data, fname):
    '''
    thin wrapper around python's json write, for consistency of interface
    '''

    if isinstance(fname, dict) and isinstance(data, str):
        # arguments were switched: accept anyway
        tmp   = data
        data  = fname
        fname = tmp

    if not data:
        # we don't have a logger to report :-/
        return

    str_data = to_json(data)
    assert str_data, 'failed to serialize data'

    dirname = os.path.dirname(fname) or '.'

    _, t_name = tempfile.mkstemp(dir=dirname)
    with open(t_name, 'w') as f_out:
        f_out.write('%s\n' % str_data)

    os.rename(t_name, fname)


# ------------------------------------------------------------------------------
#
def dumps_json(data):
    '''
    thin wrapper around python's json write, for consistency of interface

    '''

    return to_json(data)


# ------------------------------------------------------------------------------
#
def parse_json(json_str, filter_comments=True):
    '''
    Comment lines in the form of

        # some json data or text

    are stripped from json before parsing
    '''

    if filter_comments:
        json_str = '\n'.join([re.sub(r'^\s*#.*$', '', line)
                         for line in json_str.split('\n')])

    return from_json(json_str)


# ------------------------------------------------------------------------------
#
def parse_json_str(json_str):
    '''
    same as parse_json, but converts unicode strings to simple strings
    '''

    return as_string(parse_json(json_str))


# ------------------------------------------------------------------------------




# pylint: disable=protected-access

import os
import csv
import time

from .ids     import get_radical_base
from .misc    import as_string, as_list, ru_open
from .misc    import get_env_ns      as ru_get_env_ns
from .host    import get_hostname    as ru_get_hostname
from .host    import get_hostip      as ru_get_hostip
from .threads import get_thread_name as ru_get_thread_name
from .config  import DefaultConfig
from .atfork  import atfork
from .shell   import sh_callout
from .which   import which
from .modules import import_module


# ------------------------------------------------------------------------------
#
# We store profiles in CSV formatted files.
# The CSV field names are defined here:
#
TIME         = 0  # time of event (float, seconds since epoch)  mandatory
EVENT        = 1  # event ID (string)                           mandatory
COMP         = 2  # component which recorded the event          mandatory
TID          = 3  # uid of thread involved                      optional
UID          = 4  # uid of entity involved                      optional
STATE        = 5  # state of entity involved                    optional
MSG          = 6  # message describing the event                optional
ENTITY       = 7  # type of entity involved                     optional
PROF_KEY_MAX = 8  # iteration helper: `for _ in range(PROF_KEY_MAX):`

# Note that `ENTITY` is not written to the profile, but rather derived from the
# UID when reading the profiles.

# A previous incarnation of this class stored CSVs with the following columns:
#
# TIME       = 0  # time of event (float, seconds since epoch)  mandatory
# COMP       = 2  # component which recorded the event          mandatory
# TID        = 3  # uid of thread involved                      optional
# UID        = 4  # uid of entity involved                      optional
# STATE      = 5  # state of entity involved                    optional
# EVENT      = 1  # event ID (string)                           mandatory
# MSG        = 6  # message describing the event                optional

# ------------------------------------------------------------------------------
#
# when recombining profiles, we will get one NTP sync offset per profile, and
# thus potentially multiple such offsets per host.  If those differ more than
# a certain value (float, in seconds) from each other, we print a warning:
#
NTP_DIFF_WARN_LIMIT = 1.0

# syncing with the NTP host is expensive, so we only do it once in a while and
# cache the result.  We use a disk cache which is valid for 1 minute
NTP_CACHE_TIMEOUT = 60  # disk cache is valid for 60 seconds

# maximum field size allowed by the csv parser.  The larger the number of
# entities in the profile, the larger the size of the filed required by the
# csv parser. We assume a 64bit C long.
CSV_FIELD_SIZE_LIMIT = 9223372036854775807


# ------------------------------------------------------------------------------
#
def _sync_ntp():

    # read from disk cache
    try:
        with ru_open('%s/ntp.cache' % get_radical_base('utils'), 'r') as fin:
            data  = as_string(fin.read()).split()
            t_sys = float(data[0])
            t_ntp = float(data[1])

    except:
        t_sys = None
        t_ntp = None


    # if disc cache is empty or old
    t_now = time.time()
    if t_sys is None or t_now - t_sys > NTP_CACHE_TIMEOUT:

        # refresh data
        import ntplib                                    # pylint: disable=E0401

        ntp_host = os.environ.get('RADICAL_UTILS_NTPHOST','0.pool.ntp.org')

        t_one = time.time()
        response = ntplib.NTPClient().request(ntp_host, timeout=1)
        t_two = time.time()

        t_sys = (t_one + t_two) / 2.0
        t_ntp = response.tx_time

        with ru_open('%s/ntp.cache' % get_radical_base('utils'), 'w') as fout:
            fout.write('%f\n%f\n' % (t_sys, t_ntp))

    # correct both time stamps by current time
    t_cor  = time.time() - t_sys
    t_sys += t_cor
    t_ntp += t_cor

    return t_sys, t_ntp


# ------------------------------------------------------------------------------
#
# the profiler is not using threads and is generally threadsafe (all write ops
# should be atomic) - but alas Python threadlocks I/O streams, and those locks
# can still deadlock after fork:
#
#   - https://bugs.python.org/issue6721
#   - https://bugs.python.org/issue40399
#
# We thus have to close/reopen the prof file handle after fork.  This creates
# a bit of a mess as we now have to maintain a global list of profiler instances
# to clean up after fork... :-/
#
_profilers = list()


def _atfork_prepare():
    pass


def _atfork_parent():
    pass


def _atfork_child():
    for prof, fname in _profilers:
        prof._handle = ru_open(fname, 'a', buffering=1024)


atfork(_atfork_prepare, _atfork_parent, _atfork_child)


# ------------------------------------------------------------------------------
#
class Profiler(object):
    '''
    This class is really just a persistent file handle with a convenience call
    (prof()) to write lines timestamped events.  Any profiling intelligence must
    be applied when reading and evaluating the created profiles.  the following
    fields are defined for each event:

        time : mandatory, float,  time in seconds since epoch
        event: mandatory, string, short, unique name of event to be recorded
        comp : optional,  string, name of component where the event originates
        tid  : optional,  string, current thread id (name)
        uid  : optional,  string, ID of entity involved (when available)
        state: optional,  string, state of entity involved, if applicable
        msg  : optional,  string, free for message describing the event

    Strings MUST NOT contain commas.  Otherwise they are encouraged to be formed
    as `[a-z][0-9a-z_.]*'. `msg` are free-form, but the inhibition of comma
    holds.  We propose to limit the sum of strings to about 256 characters -
    this will guarantee atomic writes on most OS's, w/o additional locking
    overheads.  Less than 100 charcters makes the profiles almost
    human-readable.

    The profile is enabled by setting environment variables.  For a profiler
    named `radical.utils`, the following env variables will be evaluated:

        RADICAL_UTILS_PROFILE
        RADICAL_PROFILE

    If either is present in the environemnt, the profile is enabled (the value
    of the setting is ignored).
    '''

    fields  = ['time', 'event', 'comp', 'thread', 'uid', 'state', 'msg']

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, ns=None, path=None):
        '''
        Open the file handle, sync the clock, and write timestam_zero
        '''

        ru_def = DefaultConfig()

        if not ns:
            ns = name

        # check if this profile is enabled via an env variable
        self._enabled = ru_get_env_ns('profile', ns)

        if self._enabled is None:
            self._enabled = ru_def.get('profile', 'False')

        if self._enabled.lower() in ['0', 'false', 'off']:
            self._enabled = False

        # don't open the file on disabled profilers
        if not self._enabled:
            return

        # profiler is enabled - set properties, sync time, open handle
        self._enabled = True
        self._handle  = None
        self._path    = path
        self._name    = name

        if not self._path:
            self._path = ru_def['profile_dir']

        self._ts_zero, self._ts_abs, self._ts_mode = self._timestamp_init()

        try:
            os.makedirs(self._path)

        except OSError:
            pass  # already exists


    # --------------------------------------------------------------------------
    #
    def _open(self):

        if not self._enabled:
            return

        if self._handle:
            return

        # we set `buffering` to `1` to force line buffering.  That is not idea
        # performance wise - but will not do an `fsync()` after writes, so OS
        # level buffering should still apply.  This is supposed to shield
        # against incomplete profiles.
        fname = '%s/%s.prof' % (self._path, self._name)
        self._handle = ru_open(fname, 'a', buffering=1024)

        # register for cleanup after fork
        _profilers.append([self, fname])

        # write header and time normalization info
        self._handle.write('#%s\n' % (','.join(Profiler.fields)))
        self._handle.write('%.7f,%s,%s,%s,%s,%s,%s\n' %
                       (self.timestamp(), 'sync_abs', self._name,
                        ru_get_thread_name(), '', '',
                        '%s:%s:%s:%s:%s' % (ru_get_hostname(),
                                            ru_get_hostip(),
                                            self._ts_zero,
                                            self._ts_abs,
                                            self._ts_mode)))


    # --------------------------------------------------------------------------
    #
    def __del__(self):

      # self.close()
        pass


    # --------------------------------------------------------------------------
    #
    @property
    def enabled(self):

        return self._enabled


    @property
    def path(self):

        return self._path


    # --------------------------------------------------------------------------
    #
    def enable(self):  self._enabled = True
    def disable(self): self._enabled = False


    # --------------------------------------------------------------------------
    #
    def close(self):

        try:
            if not self._enabled:
                return

            if self._enabled:
                self.prof('END')
                self.flush()
                self._handle.close()
                self._handle  = None
                self._enabled = False

        except:
            pass


    # --------------------------------------------------------------------------
    #
    def flush(self, verbose=False):

        if not self._enabled:
            return

        # see https://docs.python.org/2/library/stdtypes.html#file.flush
        self._handle.flush()
        os.fsync(self._handle.fileno())


    # --------------------------------------------------------------------------
    #
    # FIXME: reorder args to reflect tupleorder (breaks API)
    #
    def prof(self, event, uid=None, state=None, msg=None, ts=None, comp=None,
                   tid=None):


        if not self._enabled:
            return

        self._open()

        if ts    is None: ts    = self.timestamp()
        if comp  is None: comp  = self._name
        if tid   is None: tid   = ru_get_thread_name()
        if uid   is None: uid   = ''
        if state is None: state = ''
        if msg   is None: msg   = ''

        # if uid is a list, then recursively call self.prof for each uid given
        for _uid in as_list(uid):

            data = '%.7f,%s,%s,%s,%s,%s,%s\n' \
                    % (ts, event, comp, tid, _uid, state, msg)
            self._handle.write(data)
            self._handle.flush()


    # --------------------------------------------------------------------------
    #
    def _timestamp_init(self):
        '''
        return a tuple of [system time, absolute time]
        '''

        # retrieve absolute timestamp from an external source
        #
        # We first try to contact a network time service for a timestamp, if
        # that fails we use the current system time.
        try:
            ts_sys, ts_ntp = _sync_ntp()
            return [ts_sys, ts_ntp, 'ntp']

        except:
            pass

        # on any errors, we fall back to system time
        t = time.time()
        return [t,t, 'sys']


    # --------------------------------------------------------------------------
    #
    def timestamp(self):

        return time.time()


# --------------------------------------------------------------------------
#
def timestamp():

    return time.time()


# ------------------------------------------------------------------------------
#
def read_profiles(profiles, sid=None, efilter=None):
    '''
    We read all profiles as CSV files and parse them.  For each profile,
    we back-calculate global time (epoch) from the synch timestamps.

    The caller can provide a filter of the following structure::

        filter = {ru.EVENT: ['event 1', 'event 2', ...],
                  ru.MSG  : ['msg 1',   'msg 2',   ...],
                  ...
                 }

    Filters apply on *substring* matches!
    '''

    legacy = os.environ.get('RADICAL_ANALYTICS_LEGACY_PROFILES', False)

    if legacy and legacy.lower() not in ['no', 'false']:
        legacy = True
    else:
        legacy = False

    # set the maximum field size allowed by the csv parser
    csv.field_size_limit(CSV_FIELD_SIZE_LIMIT)

  # import resource
  # print('max RSS       : %20d MB' % (resource.getrusage(1)[2]/(1024)))

    # FIXME: we correct one pesky profile entry, which is exactly 1.000 in an
    #        otherwise ntp-aligned profile - see [1].  In this case we use the
    #        previous timestamp (if available)
    #
    #    [1] https://github.com/radical-cybertools/radical.pilot/issues/1117

    if not efilter:
        efilter = dict()

    ret     = dict()
    last    = list()
    skipped = 0

    for prof in profiles:

        with ru_open(prof, 'r') as csvfile:

            ret[prof] = list()
            reader    = csv.reader(csvfile)

            try:
                for raw in reader:

                    # we keep the raw data around for error checks
                    row = list(raw)

                  # if 'bootstrap_1' in row:
                  #     print()
                  #     print(row)

                    # skip header
                    if row[TIME].startswith('#'):
                        skipped += 1
                        continue

                    # make room in the row for entity type etc.
                    row.extend([None] * (PROF_KEY_MAX - len(row)))

                    row[TIME] = float(row[TIME])

                    # we derive entity type from the uid -- but funnel
                    # some cases into 'session' as a catch-all type
                    uid = row[UID]
                    if uid:
                        row[ENTITY] = uid.split('.',1)[0]
                    else:
                        row[ENTITY] = 'session'
                        row[UID]    = sid

                    # we should have no unset (ie. None) fields left - otherwise
                    # the profile was likely not correctly closed.
                    if None in row:
                        if legacy:
                            comp, tid = row[1].split(':', 1)
                            new_row = [None] * PROF_KEY_MAX
                            new_row[TIME        ] = row[0]
                            new_row[EVENT       ] = row[4]
                            new_row[COMP        ] = comp
                            new_row[TID         ] = tid
                            new_row[UID         ] = row[2]
                            new_row[STATE       ] = row[3]
                            new_row[MSG         ] = row[5]

                            uid = new_row[UID]
                            if uid:
                                new_row[ENTITY] = uid.split('.',1)[0]
                            else:
                                new_row[ENTITY] = 'session'
                                new_row[UID]    = sid

                            row = new_row

                    if None in row:
                        print('row invalid [%s]: %s' % (prof, raw))
                        continue
                      # raise ValueError('row invalid [%s]: %s' % (prof, row))

                    # apply the filter.  We do that after adding the entity
                    # field above, as the filter might also apply to that.
                    skip = False
                    for field, pats in efilter.items():
                        for pattern in pats:
                            if row[field] in pattern:
                                skip = True
                                break
                        if skip:
                            continue

                    # fix rp issue 1117 (see FIXME above)
                    if row[TIME] == 1.0 and last:
                        row[TIME] = last[TIME]

                    if not skip:
                        ret[prof].append(row)

                    last = row

                  # print(' --- %-30s -- %-30s ' % (row[STATE], row[MSG]))
                  # if 'bootstrap_1' in row:
                  #     print(row)
                  #     print()
                  #     print('TIME    : %s' % row[TIME  ])
                  #     print('EVENT   : %s' % row[EVENT ])
                  #     print('COMP    : %s' % row[COMP  ])
                  #     print('TID     : %s' % row[TID   ])
                  #     print('UID     : %s' % row[UID   ])
                  #     print('STATE   : %s' % row[STATE ])
                  #     print('ENTITY  : %s' % row[ENTITY])
                  #     print('MSG     : %s' % row[MSG   ])

            except:
                raise
              # print('skip remainder of %s' % prof)
              # continue

    return ret


# ------------------------------------------------------------------------------
#
def combine_profiles(profs):
    '''
    We merge all profiles and sort by time.

    This routine expects all profiles to have a synchronization time stamp.
    Two kinds of sync timestamps are supported: absolute (`sync_abs`) and
    relative (`sync_rel`).

    Time syncing is done based on 'sync_abs' timestamps.  We expect one such
    absolute timestamp to be available per host (the first profile entry will
    contain host information).  All timestamps from the same host will be
    corrected by the respectively determined NTP offset.  We define an
    'accuracy' measure which is the maximum difference of clock correction
    offsets across all hosts.

    The `sync_rel` timestamps are expected to occur in pairs, one for a profile
    with no other sync timestamp, and one profile which has a `sync_abs`
    timestamp.  In that case, the time correction from the latter is transfered
    to the former (the two time stamps are considered to have been written at
    the exact same time).

    The method returnes the combined profile and accuracy, as tuple.
    '''

    syncs    = dict()  # profiles which have relative time refs
    t_host   = dict()  # time offset per host
    p_glob   = list()  # global profile
    t_min    = None    # absolute starting point of profiled session
    c_end    = 0       # counter for profile closing tag
    accuracy = 0       # max uncorrected clock deviation

    if len(profs) == 1:
        return list(profs.values())[0], accuracy

    # first get all absolute and relative timestamp sync from the profiles,
    # for all hosts
    for pname, prof in profs.items():

        sync_abs = list()
        sync_rel = list()

        syncs[pname] = {'rel' : sync_rel,
                        'abs' : sync_abs}

        if not len(prof):
          # print('empty        %s' % pname)
            continue

        for entry in prof:
            if entry[EVENT] == 'sync_abs': sync_abs.append(entry)
            if entry[EVENT] == 'sync_rel': sync_rel.append(entry)

        # we can have any number of sync_rel's - but if we find none, we expect
        # a sync_abs
        if not sync_rel and not sync_abs:
          # print('unsynced     %s' % pname)
            continue

        syncs[pname] = {'rel' : sync_rel,
                        'abs' : sync_abs}

  # for pname, prof in profs.items():
  #     if prof:
  #         print('check        %-100s: %s' % (pname, prof[0][TIME:EVENT]))

    for pname, prof in profs.items():

        if not len(prof):
          # print('empty        %s' % pname)
            continue

        # if we have only sync_rel(s), then find the offset by the corresponding
        # sync_rel in the other profiles, and determine the offset to use.  Use
        # the first sync_rel that results in an offset, and only complain if
        # none is found.
        offset       = None
        offset_event = None
        if syncs[pname]['abs']:
            offset = 0.0

        else:
            for sync_rel in syncs[pname]['rel']:
                for _pname in syncs:

                    if _pname == pname:
                        continue

                    for _sync_rel in syncs[_pname]['rel']:
                        if _sync_rel[MSG] == sync_rel[MSG]:
                            offset        = _sync_rel[TIME] - sync_rel[TIME]
                            offset_event  = syncs[_pname]['abs'][0]

                    if offset:
                        break

                if offset:
                    break

        if offset is None:
          # print('no rel sync  %s' % pname)
            continue

      # print('sync profile %-100s : %20.3fs' % (pname, offset))
        for event in prof:
            event[TIME] += offset

        # if we have an offset event, we append it to the profile.  This
        # basically transplants an sync_abs event into a sync_rel profile
        if offset_event:
          # print('transplant sync_abs to %s: %s' % (pname, offset_event))
            prof.append(offset_event)
            syncs[pname]['abs'].append(offset_event)

    # all profiles are rel-synced here.  Now we look at sync_abs values to align
    # across hosts and to determine accuracy.
    for pname in syncs:

        for sync_abs in syncs[pname]['abs']:

            # https://github.com/radical-cybertools/radical.analytics/issues/20
            if not sync_abs[MSG] or ':' not in sync_abs[MSG]:
              # print('unsynced profile %s [%s]' % (pname, sync_abs))
                continue

            t_prof = sync_abs[TIME]

            host, ip, t_sys, t_ntp, t_mode = sync_abs[MSG].split(':')
            host_id = '%s:%s' % (host, ip)

            if t_min: t_min = min(t_min, t_prof)
            else    : t_min = t_prof

            if t_mode == 'sys':
              # print('sys synced profile (%s)' % t_mode)
                continue

            # determine the correction for the given host
            t_sys = float(t_sys)
            t_ntp = float(t_ntp)
            t_off = t_sys - t_ntp

            if  host_id in t_host and \
                t_host[host_id] != t_off:

                diff = t_off - t_host[host_id]
                accuracy = max(accuracy, diff)

                # we allow for *some* amount of inconsistency before warning
                if diff > NTP_DIFF_WARN_LIMIT:
                    print('conflicting time sync for %-45s (%15s): '
                          '%10.2f - %10.2f = %5.2f'
                        % (pname.split('/')[-1], host_id, t_off,
                           t_host[host_id], diff))
                    continue

            t_host[host_id] = t_off


    unsynced = set()
    # now that we can align clocks for all hosts, apply that correction to all
    # profiles
    for pname, prof in profs.items():

        if not len(prof):
          # print('empty prof: %s' % pname)
            continue

        if not syncs[pname]['abs']:

            t_off = 0.0

        else:

            sync_abs = syncs[pname]['abs'][0]

            host, ip, _, _, _ = sync_abs[MSG].split(':')
            host_id = '%s:%s' % (host, ip)
            if host_id in t_host:
                t_off = t_host[host_id]
            else:
                unsynced.add(host_id)
                t_off = 0.0

            t_0 = sync_abs[TIME]
            t_0 -= t_min

        # correct profile timestamps
        for row in prof:

            row[TIME] -= t_min
            row[TIME] -= t_off

          # print(row[EVENT],)
            # count closing entries
            if row[EVENT] == 'END':
                c_end += 1

        # add profile to global one
        p_glob += prof

      # if prof:
      #     print('check        %-100s: %s' % (pname, prof[0][TIME:EVENT]))

        # Check for proper closure of profiling files
      # if c_end == 0:
      #     print('WARNING: profile "%s" not correctly closed.' % pname)
      # elif c_end > 1:
      #     print('WARNING: profile "%s" closed %d times.' % (pname, c_end))

    # sort by time and return
    p_glob = sorted(p_glob[:], key=lambda k: k[TIME])

  # print('check        %-100s: %s' % ('t_min', p_glob[0][TIME]))
  # print('check        %-100s: %s' % ('t_max', p_glob[-1][TIME]))
    return p_glob, accuracy


# ------------------------------------------------------------------------------
#
def clean_profile(profile, sid, state_final=None, state_canceled=None):
    '''
    This method will prepare a profile for consumption in radical.analytics.
    It performs the following actions:

      - makes sure all events have a `ename` entry
      - remove all state transitions to `CANCELLED` if a different final state
        is encountered for the same uid
      - assignes the session uid to all events without uid
      - makes sure that state transitions have an `ename` set to `state`
    '''

    entities = dict()  # things which have a uid

    if not state_final:
        state_final = []
    elif not isinstance(state_final, list):
        state_final = [state_final]

    for event in profile:

        uid   = event[UID  ]
        state = event[STATE]
        name  = event[EVENT]

        # we derive entity_type from the uid -- but funnel
        # some cases into the session
        if uid:
            event[ENTITY] = uid.split('.',1)[0]
        else:
            event[ENTITY] = 'session'
            event[UID]    = sid
            uid = sid

        if uid not in entities:
            entities[uid] = dict()
            entities[uid]['states'] = dict()
            entities[uid]['events'] = list()

        if name == 'advance':

            # this is a state progression
            assert state, 'cannot advance w/o state'
            assert uid,   'cannot advance w/o uid'

            # this is a state transition event
            event[EVENT] = 'state'

            skip = False
            if state in state_final and state != state_canceled:

                # a final state other than CANCELED will cancel any previous
                # CANCELED state.
                if  state_canceled and \
                    state_canceled in entities[uid]['states']:
                    del entities[uid]['states'][state_canceled]

                # vice-versa, we will not add CANCELED if a final
                # state already exists:
                if  state_canceled and \
                    state_canceled == state:
                    if any([s in entities[uid]['states']
                              for s in state_final]):
                        skip = True
                        continue

            if state in entities[uid]['states']:
                # ignore duplicated recordings of state transitions
                skip = True
                continue
              # raise ValueError('double state (%s) for %s' % (state, uid))

            if not skip:
                entities[uid]['states'][state] = event

        entities[uid]['events'].append(event)


    # we have evaluated, cleaned and sorted all events -- now we recreate
    # a clean profile out of them
    ret = list()
    for entity in list(entities.values()):
        ret += entity['events']

    # sort by time and return
    ret = sorted(ret[:], key=lambda k: k[TIME])

    return ret


# ------------------------------------------------------------------------------
#
def event_to_label(event):

    if event[EVENT] == 'state':
        return event[STATE]
    else:
        return event[EVENT]


# ------------------------------------------------------------------------------
#
class Yappi(object):

    def __init__(self, name, method='wall', verbose=False):
        self._name  = name
        self._verb  = verbose
        self._yappi = None
        try:
            self._yappi = import_module('yappi')
            self._yappi.set_clock_type(method)
        except:
            pass

    def __enter__(self):
        if self._yappi:
            self._yappi.start(builtins=True)

    def __exit__(self, etype, value, traceback):

        if self._yappi:

            if self._verb:
                self._yappi.get_func_stats().print_all()
                self._yappi.get_thread_stats().print_all()

            fstats = self._yappi.get_func_stats()
            pstats = self._yappi.convert2pstats(fstats)
            pstats.dump_stats('%s.pstats' % self._name)

            if which('gprof2dot'):

                cmd  = 'gprof2dot -e 0.00 -n 0.12 --skew=0.3'
                cmd += ' --node-label="total-time" -f pstats %s.pstats' % self._name
                cmd += ' | dot -Tpng -o %s.png' % self._name

                out, err, ret = sh_callout(cmd, shell=True)
                if ret:
                    print('gprof2dot failed: %s' % err)


# ------------------------------------------------------------------------------

